'''
Given a complex s21 sweep, returns the data fit to the resonator model and the
resonator parameters and provides tools for plotting the fit results.

Based on equation 11 from Kahlil et al. and adapted from Columbia KIDs open
source analysis code.
'''
import numpy as np
from lmfit import Model
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

def linear_resonator(f, f_0, Q, Q_e_real, Q_e_imag):
    '''
    Function for a resonator with asymmetry parameterized by the imaginary
    part of ``Q_e``. The real part of ``Q_e`` is what we typically refer to as
    the coupled Q, ``Q_c``.
    '''
    Q_e = Q_e_real + 1j*Q_e_imag
    return (1 - (Q * Q_e**(-1) /(1 + 2j * Q * (f - f_0) / f_0) ) )

def cable_delay(f, delay, phi, f_min):
    '''
    Function implements a time delay (phase variation linear with frequency).
    '''
    return np.exp(1j * (-2 * np.pi * (f - f_min) * delay + phi))

def general_cable(f, delay, phi, f_min, A_mag, A_slope):
    '''
    Function implements a time delay (phase variation linear with frequency) and
    attenuation slope characterizing a background RF cable transfer function.
    '''
    phase_term =  cable_delay(f,delay,phi,f_min)
    magnitude_term = ((f-f_min)*A_slope + 1)* A_mag
    return magnitude_term*phase_term

def resonator_cable(f, f_0, Q, Q_e_real, Q_e_imag, delay, phi, f_min, A_mag,
                    A_slope):
    '''
    Function that includes asymmetric resonator (``linear_resonator``) and cable
    transfer functions (``general_cable``). Which most closely matches our full
    measured transfer function.
    '''
    resonator_term = linear_resonator(f, f_0, Q, Q_e_real, Q_e_imag)
    cable_term = general_cable(f, delay, phi, f_min, A_mag, A_slope)
    return cable_term*resonator_term

def full_fit(freqs, real, imag):
    '''
    Fitting function that takes in frequency and real and imaginary parts of the
    transmission of a resonator (needs to trimmed down to only data for a
    single resonator) and returns fitted parameters to the ``resonator_cable``
    model.

    Args
    ----
    freqs : float ndarray
        Frequencies that line up with complex transmission data.
    real : float ndarray
        Real part of resonator complex transmission to be fit
    imag : float ndarray
        Imaginary part of resonator complex transmission to be fit.

    Returns
    -------
    result : (`lmfit.Model.ModelResult`)
        This is a class of lmfit that contains all of the fitted parameters as
        well as a number of other pieces of data and metadata related to the fit
        and some helper functions for plotting and manipulating the data in the
        object.
    '''
    s21_complex = np.vectorize(complex)(real, imag)

    #set our initial guesses
    argmin_s21 = np.abs(s21_complex).argmin()
    fmin = freqs.min()
    fmax = freqs.max()
    f_0_guess = freqs[argmin_s21]
    Q_min = 0.1 * (f_0_guess / (fmax - fmin))
    delta_f = np.diff(freqs)
    min_delta_f = delta_f[delta_f > 0].min()
    Q_max = f_0_guess / min_delta_f
    Q_guess = np.sqrt(Q_min * Q_max)
    s21_min = np.abs(s21_complex[argmin_s21])
    s21_max = np.abs(s21_complex).max()
    Q_e_real_guess = Q_guess / (1 - s21_min / s21_max)
    A_slope, A_offset = np.polyfit(freqs - fmin, np.abs(s21_complex), 1)
    A_mag = A_offset
    A_mag_slope = A_slope / A_mag
    phi_slope, phi_offset = np.polyfit(freqs - fmin, np.unwrap(np.angle(s21_complex)), 1)
    delay = -phi_slope / (2 * np.pi)

    #make our model
    totalmodel = Model(resonator_cable)
    params = totalmodel.make_params(f_0=f_0_guess,
                                Q=Q_guess,
                                Q_e_real=Q_e_real_guess,
                                Q_e_imag=0,
                                delay=delay,
                                phi=phi_offset,
                                f_min=fmin,
                                A_mag=A_mag,
                                A_slope=A_mag_slope)
    #set some bounds
    params['f_0'].set(min=freqs.min(), max=freqs.max())
    params['Q'].set(min=Q_min, max=Q_max)
    params['Q_e_real'].set(min=1, max=1e7)
    params['Q_e_imag'].set(min=-1e7, max=1e7)
    params['phi'].set(min=phi_offset-np.pi, max=phi_offset+np.pi)

    #fit it
    result = totalmodel.fit(s21_complex, params, f=freqs)
    return result

def get_qi(Q, Q_e_real, Q_e_imag):
    '''
    Function for deriving the internal quality factor from the fitted quality
    factors (Q and Qc).

    Args
    ----
    Q : float
        Total resonator quality factor output parameter of ``full_fit``
    Q_e_real : float
        Resonator coupled quality factor output parameter of ``full_fit``
    Returns
    -------
    Qi : float
        Resonator internal quality factor.
    '''
    Qi = (Q**-1 - np.real((Q_e_real+1j*Q_e_imag)**-1))**-1
    return Qi

def get_br(Q, f_0):
    '''
    Function for deriving the resonator bandwidth from the fit results.

    Args
    ----
    Q : float
        Total resonator quality factor output parameter of ``full_fit``
    f_0 : float
        Resonance frequency output parameter of ``full_fit`` in Hz.
    Returns
    -------
    br : float
        Resonator bandwidth in Hz.
    '''
    br = f_0/Q
    return br

def reduced_chi_squared(ydata, ymod, n_param=9, sd=None):
    '''
    Reduced chi squared in lmfit does not return something reasonable so this
    is a handwritten function to calculate it since we want standard deviation
    to be the complex error.

    Args
    ----
    ydata : float ndarray
        complex data to calculate reduced chi squared on.
    ymod : float ndarray
        model fit result of ydata at same x-locations as ydata is sampled.
    n_param : int
        Number of parameters in the fit.
    sd : float ndarray
        standard deviation of data
    Returns
    -------
    br : float
        Resonator bandwidth in Hz.
    '''
    if sd is None:
        sdr = np.std(np.real(ydata))
        sdi = np.std(np.imag(ydata))
        sd = sdr + 1j*sdi
    chisq = np.sum((np.real(ydata) - np.real(ymod))**2/((np.real(sd))**2)) +\
            np.sum((np.imag(ydata) - np.imag(ymod))**2/((np.imag(sd))**2))
    nu=2*ydata.size-n_param     #multiply  the usual by 2 since complex
    red_chisq = chisq/nu
    return chisq, red_chisq

def fit_tune(tunefile):
    """
    Automated fitting of resonator parameters from one tuning file.

    Args
    ----
    tunefile : str, filepath
        Full directories of one tunning file.
    Returns
    -------
    dres : dict
        a dictionary containing all of the fit results for all resonances in the provided tunefile. The keys are organized as::
        
          {
            tunefile: str, filepath - full directories of one tunning file
            band: 1d int array with shape (ndets) - smurf band for each detector
            channels : 1d int array with shape (ndets) - Smurf channel for each
                       detector. If unassigned, set to -1
            res_index: 1d int array with shape (ndets) -- Index of the resonator within each band
            res_freqs: 1d int array of resonance freqs (as is found in the tunefile)
            model_params : dictionary for 9 parameters of resonator_cable 
                f_0 : 1d array with shape (ndets)
                Q : 1d array with shape (ndets)
                Q_e_real : 1d array with shape (ndets)
                Q_e_imag: 1d array with shape (ndets)
                delay: 1d array with shape (ndets)
                phi: 1d array with shape (ndets)
                f_min: 1d array with shape (ndets)
                A_mag: 1d array with shape (ndets)
                A_slope: 1d array with shape (ndets)
            derived_params : dictionary for derived parameters
                Qi: 1d array with shape (ndets)
                br: 1d array with shape (ndets)
                depth: 1d array with shape (ndets)
            find_freq_ctime : 1d string array with shape (ndets) - ctime find_freq was taken
            S21: 2d array (ndets, nsamps) of measured S21
            scan_freqs: 2d array (ndets, ndsamps) of freqs used for S21 scan
            chi2 : 2d float array of shape (ndets x 2) - chi-squared goodness of fit
            sid : str, session ID
          }
    """
    dres = {}
    data = np.load(tunefile,allow_pickle=True).item()

    dres['tunefile'] = tunefile
    model_params = {'f_0': [], 'Q': [], 'Q_e_real' : [], 'Q_e_imag': [],
                   'delay': [], 'phi': [], 'f_min': [], 'A_mag': [],
                    'A_slope': []}
    derived_params = {'Qi': [], 'br': [], 'depth': []}
    bands = []
    channels = []
    res_index = []
    chi2 = []
    sid = []
    find_freq_ctime = []
    S21s = []
    scan_freqs = []
    res_freqs = []
    
    for band in sorted(list(data.keys())):
        if 'resonances' in list(data[band].keys()):
            for idx in list(data[band]['resonances'].keys()):
                scan=data[band]['resonances'][idx]
                f=scan['freq_eta_scan']
                if (band > 3) & (np.mean(f) > 6000):
                    f-=2000
                S21=scan['resp_eta_scan']
                result=full_fit(f,S21.real,S21.imag)

                # Need to check if this plays well with being in a dict/pickling
                S21_mod = result.best_fit.real+1j*result.best_fit.imag                
                
                bands.append(band)
                channels.append(scan['channel'])
                res_index.append(idx)
                res_freqs.append(scan['freq'])
                chi2.append(reduced_chi_squared(S21, S21_mod))
                find_freq_ctime.append(data[band]['find_freq']['timestamp'][0])
                S21s.append(S21)
                scan_freqs.append(f)

                model_params['f_0'].append(result.best_values['f_0'])
                model_params['Q'].append(result.best_values['Q'])
                model_params['Q_e_real'].append(result.best_values['Q_e_real'])
                model_params['Q_e_imag'].append(result.best_values['Q_e_imag'])
                model_params['delay'].append(result.best_values['delay'])
                model_params['phi'].append(result.best_values['phi'])
                model_params['f_min'].append(result.best_values['f_min'])
                model_params['A_mag'].append(result.best_values['A_mag'])
                model_params['A_slope'].append(result.best_values['A_slope'])
                
                Qi = get_qi(result.best_values['Q'], result.best_values['Q_e_real'],
                            result.best_values['Q_e_imag'])
                br = get_br(result.best_values['Q'], result.best_values['f_0'])
                depth = np.abs(S21_mod).max()/np.abs(S21_mod).min()
                derived_params['Qi'].append(Qi)
                derived_params['br'].append(br) 
                derived_params['depth'].append(depth)
                
    bands = np.array(bands)
    channels = np.array(channels)
    res_index = np.array(res_index)
    S21s = np.array(S21s)
    scan_freqs = np.array(scan_freqs)
    chi2 = np.array(chi2)
    sid = np.array(sid)
    for par in model_params:
        model_params[par] = np.array(model_params[par])
    for par in derived_params:
        derived_params[par] = np.array(derived_params[par])

    dres['bands'] = bands
    dres['channels'] = channels
    dres['res_index'] = res_index
    dres['res_freqs'] = res_freqs
    dres['sid'] = sid
    dres['S21'] = S21s
    dres['scan_freqs'] = scan_freqs
    dres['chi2'] = chi2
    dres['model_params'] = model_params
    dres['derived_params'] = derived_params
    return dres

def get_resfit_plot_txt(resfit_dict, band, rix):
    '''
    Function to assemble some key fit information out of the fit dictionary
    into a text block for adding to plots.

    Args
    ----
    resfit_dict : dict
        Dictionary with fit results output from ``fit_tune``
    band : int
        Smurf band of channel to get plot text for.
    rix : int
        Resonator index in tunefile (sorted by frequency order of setup_notches
        channels) of channel to get plot text for.

    Returns
    -------
    text : str
        text block to add to resonator fit channel plot.
    '''
#     Get index of band, rix pair provided by user
    idx = np.where(
        (resfit_dict['res_index'] == rix)
        & (resfit_dict['bands'] == band)
    )[0][0]
    # idx = int(np.where((resfit_dict['res_index'][:,0] == band) \
    #              & (resfit_dict['res_index'][:,1] == rix))[0])
            
    mparams = resfit_dict['model_params']
    dparams = resfit_dict['derived_params']
    channel = resfit_dict['channels'][idx]
    chi2 = resfit_dict['chi2'][idx]
    if channel == -1:
        text = f'Channel unassigned'
    else:
        text = f'Band {band} Channel {channel}'
    text += '\n$f_r$: '+ f"{np.round(mparams['f_0'][idx],1)} MHz"
    text += '\n$Q_i$: '+ f"{int(dparams['Qi'][idx])}"
    text += '\nBW: '+ f"{int(dparams['br'][idx]*1e3)} kHz"
#     text += '\nfit $\chi^2$: ' + f"{np.round(chi2,3)}"
    return text

def plot_channel_fit(fit_dict, idx):
    '''
    Function for plotting single channel eta_scan data from a tunefile with
    a fit to an asymmetric resonator model ``resonator_cable``.

    Args
    ----
    fit_dict : dict
        fit results dictionary from ``fit_tune``
    band : int
        smurf band of resonator to plot
    channel : int
        smurf channel of resonator to plot
    '''

    rix = fit_dict['res_index'][idx]
    band = fit_dict['bands'][idx]
    
    freqs = fit_dict['scan_freqs'][idx]
    res_freq = fit_dict['res_freqs'][idx]
    freqs_plot = 1e3*(freqs - res_freq)
    resp_data = fit_dict['S21'][idx]
    
    params = {}
    for p_opt in fit_dict['model_params']:
        params[p_opt] = fit_dict['model_params'][p_opt][idx]
    
    resp_model = resonator_cable(freqs,**params)
    fr_idx = np.argmin(np.abs(freqs - fit_dict['model_params']['f_0'][idx]))

    fig = plt.figure(figsize = (12,6),constrained_layout=True)
    gs = GridSpec(2, 6, figure=fig)
    ax1 = fig.add_subplot(gs[:1,0:2])
    ax1.plot(freqs_plot,20*np.log10(np.abs(resp_data)),
             'C0o',label = 'Data')
    ax1.plot(freqs_plot,20*np.log10(np.abs(resp_model)),
             'C1-',label = 'Fit')
    ax1.plot(freqs_plot[fr_idx],20*np.log10(np.abs(resp_data[fr_idx])),
             'rx',ms = 12,label = '$f_r$ - fit')
    ax1.set_xlabel('Offset $(f-f_{min})$ [kHz]')
    ax1.set_ylabel('$|S_{21}|$ [dB]')
    ax1.legend(loc = 'lower left',fontsize = 12)
    ax2 = fig.add_subplot(gs[1:,0:2])
    ax2.plot(freqs_plot,np.rad2deg(np.unwrap(np.angle(resp_data))),'C0o')
    ax2.plot(freqs_plot,np.rad2deg(np.unwrap(np.angle(resp_model))),'C1-')
    ax2.plot(freqs_plot[fr_idx],np.rad2deg(np.unwrap(np.angle(resp_data))[fr_idx]),
             'rx',ms = 12)
    ax2.set_xlabel('Offset $(f-f_r)$ [kHz]')
    ax2.set_ylabel('Phase $S_{21}$ [$^{\\circ}$]')
    ax3 = fig.add_subplot(gs[:,2:])
    ax3.plot(np.real(resp_data),np.imag(resp_data),'C0o')
    ax3.plot(np.real(resp_model),np.imag(resp_model),'C1-')
    ax3.plot(np.real(resp_model[fr_idx]),np.imag(resp_model[fr_idx]),
             'rx',ms = 12)
    ax3.set_xlabel('Re($S_{21}$) [I]')
    ax3.set_ylabel('Im($S_{21}$) [Q]')
    mrange = 1.1*np.max(np.concatenate((np.abs(np.real(resp_data)),
                                   np.abs(np.imag(resp_data)),
                                   np.abs(np.real(resp_model)),
                                   np.abs(np.imag(resp_model)))))
    ax3.set_xlim(-mrange,mrange)
    ax3.set_ylim(-mrange,mrange)
    ax3.axhline(0,color = 'k')
    ax3.axvline(0,color = 'k')

    restext = get_resfit_plot_txt(fit_dict, band, rix)
    ax3.text(0.025,0.05,restext,
             bbox=dict(facecolor='wheat',
                       alpha=0.5,
                      boxstyle="round",),
            transform=ax3.transAxes)
    return fig, np.array([ax1, ax2, ax3])

def plot_fit_summary(fit_dict, plot_style=None, quantile=0.98):
    '''
    Function for plotting full wafer eta_scan data from a tunefile.
        Plots distributions of Qi, dip depth, bandwidth, and frequency spacing
        for all resonators within specified quantile.

    Args
    ----
    fit_dict : dict
        fit results dictionary from ``fit_tune``
    plot_style : dict
        keyword arguments to pass to the histogram plotting for formatting.
    quantile: float
        inner X percent to graph
    '''
    
    Qis = fit_dict['derived_params']['Qi']
    depths = fit_dict['derived_params']['depth']
    bws = fit_dict['derived_params']['br']
    frs = fit_dict['model_params']['f_0']
    seps = np.diff(frs)

    Qi_quant = Qis[(Qis>=np.nanquantile(Qis, 1-quantile)) \
                   & (Qis<=np.nanquantile(Qis, quantile))]
    depth_quant = depths[(depths>=np.nanquantile(depths, 1-quantile)) \
                   & (depths<=np.nanquantile(depths, quantile))]
    bw_quant = bws[(bws>=np.nanquantile(bws, 1-quantile)) \
                   & (bws<=np.nanquantile(bws, quantile))]
    seps_quant = seps[(seps>=np.nanquantile(seps, 1-quantile)) \
                   & (seps<=np.nanquantile(seps, quantile))]
    
    fig, axes = plt.subplots(2, 2, figsize=(16, 8))

    if plot_style is None:
        plot_style = {'bins': 30,
                      'color': 'gold',
                      'alpha': 0.5,
                      'edgecolor': 'orange',
                      'lw': 2}
    #Qi plot
    ax = axes[0, 0]
    ax.hist(Qi_quant/1e5, **plot_style)
    Qi_med = np.median(Qi_quant/1e5)
    ax.axvline(np.median(Qi_quant/1e5),color = 'purple',
                label = f'Median: {np.round(Qi_med,2)} $\\times10^5$')
    ax.legend(loc = 'upper right')
    ax.set_xlabel('$Q_i\\times10^5$')
    ax.set_ylabel('Counts')

    #Dip depth plot
    ax = axes[0, 1]
    ax.hist(20*np.log10(depth_quant),**plot_style)
    dep_med = np.median(20*np.log10(depth_quant))
    ax.axvline(dep_med,color = 'purple',
                label = f'Median: {np.round(dep_med,2)} dB')
    ax.legend(loc = 'upper right')
    ax.set_xlabel('Dip Depth [dB]')
    ax.set_ylabel('Counts')

    #Bandwidth plot
    ax = axes[1, 0]
    ax.hist(bw_quant*1e3, **plot_style)
    bw_med = np.median(bw_quant*1e3)
    ax.axvline(bw_med,color = 'purple',
                label = f'Median: {np.round(bw_med,2)} kHz')
    ax.legend(loc = 'upper right')
    ax.set_xlabel('Bandwidth [kHZ]')
    ax.set_ylabel('Counts')

    #Frequency Separation plot
    ax = axes[1, 1]
    mean, std = np.nanmean(seps_quant), np.std(seps_quant)
    nstd = 2
    rng = (
        max(np.min(seps_quant), mean - nstd * std),
        min(np.max(seps_quant), mean + nstd * std),
    )

    ax.hist(seps, range=rng, **plot_style)
    sep_med = np.median(seps_quant)
    ax.axvline(sep_med,color = 'purple',
                label = f'Median: {np.round(sep_med,2)} MHz')
    ax.legend(loc = 'upper right')
    ax.set_xlabel('Resonator Separation [Mhz]')
    ax.set_ylabel('Counts')

    plt.tight_layout()
    return fig, axes



from sodetlib.operations import uxm_relock, tracking, uxm_setup
from sodetlib import noise
import sodetlib as sdl
import numpy as np
import time
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
from pysmurf.client.base.smurf_control import SmurfControl


def find_min_total_atten(S, band, atten_offset=2):
    """
    Finds the minimum total atten level (UC + DC) required for the ADCs not to
    be saturated. This assumes tones are already enabled.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    band : int
        Band number
    atten_offset : int
        Int to add to total attens to try to avoid ADC saturation.
    """
    uc, dc = 0, 0
    S.set_att_uc(band, 0)
    S.set_att_dc(band, 0)

    for uc in range(0, 31, 2):
        S.set_att_uc(band, uc)
        if not S.check_adc_saturation(band):
            return min(uc + dc + atten_offset, 60)
        time.sleep(0.05)

    for dc in range(0, 31, 2):
        S.set_att_dc(band, dc)
        if not S.check_adc_saturation(band):
            return min(uc + dc + atten_offset, 60)
        time.sleep(0.05)

    return False

@sdl.set_action()
def optimize_band_atten(S: SmurfControl, cfg, band, meas_time=30, 
    total_att=None, ucs=None, show_pb=False, bgs=None, full_tune=False):
    """
    This function will optimize values for uc and dc attenuators. First, it
    will find the minimum total attenuation required to avoid saturating ADCs.
    Then it loops through values of UC attenuators (keeping the total atten
    constant), runs gradient descent, tracking setup, and the takes noise.
    It will chose the optimal attenuation to minimize the white noise level.
    After this is run, you should re-run grad-descent and tracking functions
    before taking data.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    meas_time : float
        Duration (sec) for noise measurement
    total_atts : list
        Total attenuations for each specified band. If this is not passed,
        it will calculate this automatically
    ucs : list
        List of UC attenuations to loop over
    show_pb : bool
        If True, will display a progress bar.
    bgs : list
        List of bias lines to overbias. This must at least contain detectors on
        the band that is being optimized. If None is specified, will just
        overbias all detectors at each step.
    full_tune : bool
        If True, will run ``setup_tune``, which performs find freq and setup
        otches, at each step. This is the way to get the most accurate
        optimization, as eta and the resonance frequency change as you adjust
        the uc attenuation. If your uc step is small enough (1 or maybe 2), then
        the gradient descent and eta scan functions may be sufficient and you
        might not need to do a full tune at each step.
    """

    if total_att is None:
        total_att = find_min_total_atten(S, band)
        S.log(f"Total att for band {band}: {total_att}")

    if bgs is None:
        bgs = cfg.dev.exp['active_bgs']
    bgs = np.atleast_1d(bgs)

    if ucs is None:
        ucs = np.arange(0, 31, 3)

    nsteps = len(ucs)
    sids = np.zeros(nsteps)
    wls = np.full(nsteps, np.nan)

    wls_full = [None for _ in range(nsteps)]
    tunefiles = [None for _ in range(nsteps)]

    for i, uc in enumerate(tqdm(ucs, disable=not show_pb)):
        dc = max(total_att - uc, 0)
        if dc <= 30:
            S.set_att_uc(band, uc)
            S.set_att_dc(band, dc)
        else:
            continue

        S.log(f"Running UC={uc}")

        S.flux_ramp_off()
        S.set_flux_ramp_dac(0)
        for bg in bgs:
            S.set_tes_bias_bipolar(bg, 0)

        if full_tune:
            uxm_setup.setup_tune(S, cfg, bands=band, update_cfg=False)
        else:
            uxm_relock.run_grad_descent_and_eta_scan(S, cfg, bands=[band])

        tunefiles[i] = S.tune_file
        tracking.relock_tracking_setup(S, cfg, band)
        S.overbias_tes_all(bgs)
        _, res = noise.take_noise(S, cfg, meas_time, plot_band_summary=False, show_plot=False, save_plot=False)
        sids[i] = res['sid']
        wls[i] = res['band_medians'][band]
        wls_full[i] = res['noise_pars'][:, 0]

    fname = sdl.make_filename(S, f'atten_optimization_b{band}.npy')
    try:
        wls_full = np.array(wls_full)
    except:
        print('Error: not all wls_full entries are of same shape. Could be due to different number of resonators on each step')
        wls_full = None
    data = dict(
        band=band, total_att=total_att, ucs=ucs, sids=sids, wls=wls,
        wls_full=wls_full, path=fname, tunefiles=tunefiles, meta=sdl.get_metadata(S, cfg)
    )

    np.save(fname, data, allow_pickle=True)
    S.log(f"Saving {fname}")
    S.pub.register_file(fname, 'atten_optimzation', format='npy')

    fig, _ = plot_atten_optimization(data)
    fname = sdl.make_filename(S, 'atten_optimization.png', plot=True)
    S.log(f"Saving {fname}")
    fig.savefig(fname)
    S.pub.register_file(fname, 'atten_optimzation', format='png')

    return data


def plot_atten_optimization(data, ax=None, ylim=None, text_loc=(0.02, 0.03)):
    """
    Plots the results of the optimize_attens function.

    Args
    -----
    data : dict
        Results from the ``optimize_band_atten`` function
    ax : Matplotlib axis
        Axis to plot on. If None this will create a new figure.
    ylim : tuple
        ylim to set
    text_loc : tuple
        Location of the textbox
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    ax.plot(data['ucs'], data['wls'])

    imin = np.nanargmin(data['wls'])
    opt_uc = data['ucs'][imin]
    opt_dc = max(data['total_att'] - opt_uc, 0)
    ax.plot([opt_uc], [data['wls'][imin]], marker='x', color='red')
    ax.set_title(f"Band {data['band']}, total_att={data['total_att']}")
    txt = '\n'.join([
        f"Min = {data['wls'][imin]:.1f} pA/rt(Hz)",
        f"Opt (UC, DC) = ({opt_uc}, {opt_dc})"
    ])
    ax.text(*text_loc, txt, transform=ax.transAxes, 
            bbox=dict(facecolor='wheat', alpha=0.2))
    if ylim is not None:
        ax.set_ylim(*ylim)

    return fig, ax



import numpy as np
import time
from scipy.interpolate import interp1d
import sodetlib as sdl
import matplotlib.pyplot as plt
import os
from copy import deepcopy
from tqdm.auto import trange
from typing import Optional, Dict, Any
from dataclasses import dataclass, asdict

import warnings
warnings.filterwarnings('ignore')
np.seterr(all="ignore")

class IVAnalysis:
    """
    IVAnalysis is the object used to hold take_iv information and analysis
    products. When instantiating from scratch, all keyword arguments must be
    specified. (This is not true if you are loading from an existing file
    using the load function).

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    run_kwargs : dict
        Dictionary of keyword arguments passed to the take_iv function.
    sid : int
        Session id from the IV stream session
    start_times : np.ndarray
        Array of start_times of each bias point
    stop_times : np.ndarray
        Array of stop_times of each bias point

    Attributes
    -----------
    meta : dict
        Dictionary of pysmurf and sodetlib metadata during the IV.
    biases_cmd : np.ndarray
        Array of commanded biases voltages (in low-current-mode units)
    nbiases : int
        Number of bias points commanded throughout the IV
    bias_groups : np.ndarray
        Array containing the bias-groups included in this IV
    am : AxisManager
        AxisManager containing TOD from the IV (this is not saved to disk,
        but can be loaded with the _load_am function)
    nchans : int
        Number of channels included in the IV
    bands : np.ndarray
        Array of shape (nchans) containing the smurf-band of each channel
    channels : np.ndarray
        Array of shape (nchans) containing the smurf-channel of each channel
    v_bias : np.ndarray
        Array of shape (nbiases) containing the bias voltage at each bias point
        (in low-current-mode Volts)
    i_bias : np.ndarray
        Array of shape (nbiases) containing the commanded bias-current
        at each step (Amps)
    resp : np.ndarray
        Array of shape (nchans, nbiases) containing the squid response (Amps)
        at each bias point
    R : np.ndarray
        Array of shape (nchans, nbiases) containing the TES Resistance of each
        channel at each bias point
    p_tes : np.ndarray
        Array of shape (nchans, nbiases) containing the electrical power on the
        TES (W) of each channel at each bias point
    v_tes : np.ndarray
        Array of shape (nchans, nbiases) containing the voltage across the TES
        for each channel at each bias point (V)
    i_tes : np.ndarray
        Array of shape (nchans, nbiases) containing the current across the TES
        for each channel at each bias point (Amps)
    R_n : np.ndarray
        Array of shape (nchans) containing the normal resistance (Ohms) of the
        TES
    R_L : np.ndarray
        Array of shape (nchans) containing the non-TES resistance (Ohms).
        Should be shunt resistance + parasitic resistance
    p_sat : np.ndarray
        Array of shape (nchans) containing the electrical power (W) at which
        Rfrac crosses the 90% threshold (or whatever arg is passed to the
        analysis fn). Note that this is only truly the saturation power in
        the absence of optical power.
    si : np.ndarray
        Array of shape (nchans, nbiases) containing the responsivity (1/V)
        of the TES for each bias-step
    idxs : np.ndarray
        Array of shape (nchans, 3) containing:
            1. Last index of the SC branch
            2. First index off the normal branch
            3. Index at which p_tes crosses the 90% thresh
    bgmap : np.ndarray
        Array of shape (nchans) containing bias-group assignment of each
        channel in the IV. This is not calculated with the IV but pulled in
        from the device cfg, so it's important you run the take_bgmap function
        to generate the bias-group map before running the IV.
    polarity : np.ndarray
        Array of shape (nchans) containing the polarity of each channel (also
        from the bias-map file in the device cfg). This tells you if the TES
        current changes in the same direction or opposite direction of the bias
        current.
    """
    def __init__(self, S=None, cfg=None, run_kwargs=None, sid=None,
                 start_times=None, stop_times=None) -> None:

        self._S = S
        self._cfg = cfg
        self.run_kwargs = run_kwargs
        self.sid = sid
        self.start_times = start_times
        self.stop_times = stop_times
        self.am = None

        if self._S is not None:
            # Initialization stuff on initial creation with a smurf instance
            self.meta = sdl.get_metadata(S, cfg)
            self.biases_cmd = np.sort(self.run_kwargs['biases'])
            self.nbiases = len(self.biases_cmd)
            self.bias_groups = self.run_kwargs['bias_groups']

            am = self._load_am()
            self.nchans= len(am.signal)

            self.bands = am.ch_info.band
            self.channels = am.ch_info.channel
            self.v_bias = np.full((self.nbiases, ), np.nan)
            self.i_bias = np.full((self.nbiases, ), np.nan)
            self.v_thevenin = np.full((self.nbiases, ), np.nan)
            self.resp = np.full((self.nchans, self.nbiases), np.nan)
            self.R = np.full((self.nchans, self.nbiases), np.nan)
            self.p_tes = np.full((self.nchans, self.nbiases), np.nan)
            self.v_tes = np.full((self.nchans, self.nbiases), np.nan)
            self.i_tes = np.full((self.nchans, self.nbiases), np.nan)
            self.R_n = np.full((self.nchans, ), np.nan)
            self.R_L = np.full((self.nchans, ), np.nan)
            self.p_sat = np.full((self.nchans, ), np.nan)
            self.si = np.full((self.nchans, self.nbiases), np.nan)
            self.idxs = np.full((self.nchans, 3), -1, dtype=int)

            self.bgmap, self.polarity = sdl.load_bgmap(
                self.bands, self.channels, self.meta['bgmap_file']
            )

    def to_dict(self):
        saved_fields = [
            'meta', 'bands', 'channels', 'sid', 'start_times', 'stop_times',
            'run_kwargs', 'biases_cmd', 'bias_groups', 'nbiases', 'nchans',
            'bgmap', 'polarity', 'resp', 'v_bias', 'i_bias', 'R', 'R_n', 'R_L',
            'idxs', 'p_tes', 'v_tes', 'i_tes', 'p_sat', 'si', 'v_thevenin'
        ]
        return {
            field: getattr(self, field, None) for field in saved_fields
        }

    def save(self, path=None, update_cfg=False):
        data = self.to_dict()
        if path is not None:
            np.save(path, data, allow_pickle=True)
        else:
            self.filepath = sdl.validate_and_save(
                'iv_analysis.npy', data, S=self._S, cfg=self._cfg, register=True,
                make_path=True)
            if update_cfg:
                self._cfg.dev.update_experiment({'iv_file': self.filepath},
                                                update_file=True)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'IVAnalysis':
        iva = cls()
        for key, val in data.items():
            setattr(iva, key, val)

        if len(iva.start_times) == 1:
            # Data was taken before we started saving timestamps for each
            # BG separately. Convert this to the 2d array that's now expected
            # by analysis.
            iva.start_times = np.vstack([iva.start_times for _ in range(15)])
            iva.stop_times = np.vstack([iva.stop_times for _ in range(15)])

        return iva

    @classmethod
    def load(cls, path) -> 'IVAnalysis':
        return cls.from_dict(np.load(path, allow_pickle=True).item())


    def _load_am(self, arc=None):
        if self.am is None:
            if arc:
                self.am = arc.load_data(self.start_times[0],
                                        self.stop_times[-1])
            else:
                self.am = sdl.load_session(self.meta['stream_id'], self.sid)
        return self.am


def compute_psats(iva, psat_level=0.9):
    """
    Computes Psat for an IVAnalysis object. Will save results to iva.p_sat.
    This assumes i_tes, v_tes, and r_tes have already been calculated.

    Args
    ----
    iva : IVAnalysis
        IV Analysis object
    psat_level : float
        R_frac level for which Psat is defined. If 0.9, Psat will be the
        power on the TES when R_frac = 0.9.

    Returns
    -------
    p_sat : np.ndarray
        Array of length (nchan) with the p-sat computed for each channel (W)
    """
    # calculates P_sat as P_TES when Rfrac = psat_level
    # if the TES is at 90% R_n more than once, just take the last crossing
    for i in range(iva.nchans):
        if np.isnan(iva.R_n[i]):
            continue

        level = psat_level
        R = iva.R[i]
        R_n = iva.R_n[i]
        p_tes = iva.p_tes[i]
        cross_idx = np.where(R/R_n < level)[0]

        if len(cross_idx) == 0:
            iva.p_sat[i] = np.nan
            continue

        # Takes cross-index to be the last time Rfrac crosses psat_level
        cross_idx = cross_idx[-1]
        if cross_idx == 0:
            iva.p_sat[i] = np.nan
            continue

        iva.idxs[i, 2] = cross_idx
        try:
            iva.p_sat[i] = interp1d(
                R[cross_idx:cross_idx+2]/R_n,
                p_tes[cross_idx:cross_idx+2]
            )(level)
        except ValueError:
            iva.p_sat[i] = np.nan

    return iva.p_sat


def compute_si(iva):
    """
    Computes responsivity S_i for an IV analysis object. Will save results
    to iva.si. This assumes i_tes, v_tes, and r_tes have already been
    calculated.

    Args
    ----
    iva : IVAnalysis
        IV Analysis object

    Returns
    -------
    si : np.ndarray
        Array of length (nchan, nbiases) with  the responsivity as a fn of bias
        voltage for each channel (V^-1).
    """
    smooth_dist = 5
    w_len = 2 * smooth_dist + 1
    w = (1./float(w_len))*np.ones(w_len)  # window

    for i in range(iva.nchans):
        if np.isnan(iva.R_n[i]):
            continue

        # Running average
        i_tes_smooth = np.convolve(iva.i_tes[i], w, mode='same')
        v_tes_smooth = np.convolve(iva.v_tes[i], w, mode='same')
        r_tes_smooth = v_tes_smooth/i_tes_smooth

        sc_idx = iva.idxs[i, 0]
        if sc_idx == -1:
            continue
        R_L = iva.R_L[i]

        # Take derivatives
        di_tes = np.diff(i_tes_smooth)
        dv_tes = np.diff(v_tes_smooth)
        R_L_smooth = np.ones(len(r_tes_smooth-1)) * R_L
        R_L_smooth[:sc_idx] = dv_tes[:sc_idx]/di_tes[:sc_idx]
        r_tes_smooth_noStray = r_tes_smooth - R_L_smooth
        i0 = i_tes_smooth[:-1]
        r0 = r_tes_smooth_noStray[:-1]
        rL = R_L_smooth[:-1]
        beta = 0.

        # artificially setting rL to 0 for now,
        # to avoid issues in the SC branch
        # don't expect a large change, given the
        # relative size of rL to the other terms
        rL = 0

        # Responsivity estimate, derivation done here by MSF
        # https://www.overleaf.com/project/613978cb38d9d22e8550d45c
        si = -(1./(i0*r0*(2+beta)))*(1-((r0*(1+beta)+rL)/(dv_tes/di_tes)))
        si[:sc_idx] = np.nan
        iva.si[i, :-1] = si


def analyze_iv(iva, psat_level=0.9, save=False, update_cfg=False, show_pb=False):
    """
    Runs main analysis for an IVAnalysis object. This calculates the attributes
    defined in the IVAnalysis class.

    Args
    ----
    iva : IVAnalysis
        IV analysis object
    psat_level : float
        R_frac for which P_sat is defined. Defaults to 0.9
    save : bool
        If true, will save IVAnalysis object after completing
    update_cfg : bool
        If true, will update the device config with the new IV analysis
        filepath
    show_pb : bool
        If true, will display a progress bar for the IV analysis
    """
    am = iva._load_am()
    R_sh = iva.meta['R_sh']

    ts = am.primary.UnixTime / 1e9
    # Calculate phase response and bias_voltages / currents
    for i in trange(iva.nbiases, disable=(not show_pb)):
        # Start from back because analysis is easier low->high voltages
        for j, bg in enumerate(iva.bias_groups):
            t0, t1 = iva.start_times[bg, -(i+1)], iva.stop_times[bg, -(i+1)]
            chan_mask = iva.bgmap == bg
            m = (t0 < ts) & (ts < t1)
            iva.resp[chan_mask, i] = np.nanmean(am.signal[:, m][chan_mask], axis=1)
            if j == 0:
                bias_bits = np.median(am.biases[bg, m])
                iva.v_bias[i] = bias_bits * 2 * iva.meta['rtm_bit_to_volt']

    if iva.run_kwargs['high_current_mode']:
        iva.v_bias *= iva.meta['high_low_current_ratio']

    iva.i_bias = iva.v_bias / iva.meta['bias_line_resistance']

    # Convert phase to Amps
    A_per_rad = iva.meta['pA_per_phi0'] / (2*np.pi) * 1e-12
    iva.resp = (iva.resp.T * iva.polarity).T * A_per_rad

    for i in trange(iva.nchans, disable=(not show_pb)):
        d_resp = np.diff(iva.resp[i])
        dd_resp = np.diff(d_resp)
        dd_resp_abs = np.abs(dd_resp)

        # Find index of superconducting branch
        try:
            sc_idx = np.nanargmax(dd_resp_abs) + 1
        except ValueError:
            continue
        if sc_idx == 1:
            continue
        iva.idxs[i, 0] = sc_idx

        # Find index of normal branch by finding the min index after
        # sc branch. (Skips a few indices after sc branch to avoid possible
        # phase skipping)
        try:
            nb_idx = sc_idx + 1 + np.nanargmin(iva.resp[i, sc_idx+1:])
        except ValueError:
            continue
        iva.idxs[i, 1] = nb_idx
        nb_fit_idx = (iva.nbiases + nb_idx) // 2
        fin_idx = (np.isfinite(iva.i_bias[nb_fit_idx:]) &
                   np.isfinite(iva.resp[i, nb_fit_idx:]))
        try:
            norm_fit = np.polyfit(iva.i_bias[nb_fit_idx:][fin_idx],
                                  iva.resp[i, nb_fit_idx:][fin_idx], 1)
        except:
            continue
        iva.resp[i] -= norm_fit[1]  # Put resp in real current units

        fin_idx = (np.isfinite(iva.i_bias[:sc_idx]) &
                   np.isfinite(iva.resp[i, :sc_idx]))
        try:
            sc_fit = np.polyfit(iva.i_bias[:sc_idx][fin_idx],
                                iva.resp[i, :sc_idx][fin_idx], 1)
        except:
            continue

        # subtract off unphysical y-offset in superconducting branch; this
        # is probably due to an undetected phase wrap at the kink between
        # the superconducting branch and the transition, so it is
        # *probably* legitimate to remove it by hand.
        iva.resp[i, :sc_idx] -= sc_fit[1]
        sc_fit[1] = 0  # now change s.c. fit offset to 0 for plotting

        R = R_sh * (iva.i_bias/(iva.resp[i]) - 1)
        R_n = np.nanmean(R[nb_fit_idx:])
        R_L = np.nanmean(R[1:sc_idx])

        iva.v_tes[i] = iva.i_bias * R_sh * R / (R + R_sh)
        iva.i_tes[i] = iva.v_tes[i] / R
        iva.p_tes[i] = (iva.v_tes[i]**2) / R
        iva.R[i] = R
        iva.R_n[i] = R_n
        iva.R_L[i] = R_L

    compute_psats(iva, psat_level)
    compute_si(iva)

    if save:
        iva.save(update_cfg=update_cfg)

def plot_Rfracs(iva, Rn_range=(5e-3, 12e-3), bgs=None):
    """
    Plots Stacked Rfrac curves of each channel.
    """
    fig, ax = plt.subplots()
    Rfrac = (iva.R.T / iva.R_n).T
    if bgs is None:
        bgs = np.arange(12)
    bgs = np.atleast_1d(bgs)
    for i, rf in enumerate(Rfrac):
        bg = iva.bgmap[i]
        if bgs is not None:
            if bg not in bgs:
                continue

        if not Rn_range[0] < iva.R_n[i] < Rn_range[1]:
            continue

        if bg == -1:
            continue

        ax.plot(iva.v_bias, rf, alpha=0.1, color=f'C{bg}')
    ax.set(ylim=(0, 1.1))
    ax.set_xlabel("Bias Voltage (V)", fontsize=14)
    ax.set_ylabel("$R_\\mathrm{frac}$", fontsize=14)
    return fig, ax

def plot_Rn_hist(iva, range=(0, 10), text_loc=(0.05, 0.05), bbox_props=None):
    """
    Plots summary of channel normal resistances.

    Args
    -----
    iva : IVAnalysis
        IVAnalysis object
    range : tuple
        Range of histogram
    text_loc : tuple
        Location of text box in coordinate frame of axis transform
    bbox_props : dict
        Additional bbox properties to add to the textbox
    """
    fig, ax = plt.subplots()
    hist = ax.hist(iva.R_n*1000, range=range, bins=40)
    ax.axvline(np.nanmedian(iva.R_n)*1000, color='red', ls='--')

    chans_pictured = int(np.sum(hist[0]))
    txt = f"{chans_pictured} / {iva.nchans} channels pictured"
    txt += '\n' + get_plot_text(iva)
    txt += f'\nMedian: {np.nanmedian(iva.R_n)*1000:0.2f} mOhms'

    bbox = dict(facecolor='wheat', alpha=0.8)
    if bbox_props is not None:
        bbox.update(bbox_props)

    ax.text(*text_loc, txt, bbox=bbox, transform=ax.transAxes)
    ax.set_xlabel("$R_n$ (m$\\Omega$)", fontsize=14)
    return fig, ax


def plot_channel_iv(iva, rc):
    """

    Plots anlayzed IV results for a given channel.

    Args
    ----
    iva : IVAnalysis
        Analyzed IVAnalysis instance
    rc : int
        Readout channel to plot
    """
    fig, axes = plt.subplots(4, 1, figsize=(10, 10))
    for ax in axes.ravel():
        v_sc = iva.v_bias[iva.idxs[rc, 0]]
        v_norm = iva.v_bias[iva.idxs[rc, 1]]
        ax.axvspan(0, v_sc, alpha=0.1, color='C0')
        ax.axvspan(v_sc, v_norm, alpha=0.1, color='C2')
        ax.axvspan(v_norm, iva.v_bias[-1], alpha=0.1, color='C1')

    axes[0].plot(iva.v_bias, iva.i_tes[rc], color='black')
    axes[0].set_ylabel("Current (Amps)")
    axes[1].plot(iva.v_bias, iva.R[rc]*1000, color='black')
    axes[1].set_ylabel("R (mOhms)")
    axes[2].plot(iva.v_bias, iva.p_tes[rc]*1e12, color='black')
    axes[2].set_ylabel("Pbias (pW)")
    axes[3].plot(iva.v_bias, iva.si[rc]*1e-6, color='black')
    axes[3].set_ylabel("Si (1/uV)")

    b, c, bg = iva.bands[rc], iva.channels[rc], iva.bgmap[rc]
    axes[0].set_title(f"Band: {b}, Chan: {c}, BG: {bg}", fontsize=18)

    axes[-1].set_xlabel("Voltage (V)")
    return fig, axes

@dataclass
class IVConfig:
    bias_groups: Optional[list] = None
    overbias_voltage: float = 18.0
    overbias_wait: float = 5.0
    high_current_mode: bool = True
    cool_wait: float = 30.
    cool_voltage: Optional[float] = None
    biases: Optional[np.ndarray] = None
    bias_high: float = 18.0
    bias_low: float = 0.0
    bias_step: float = 0.025
    show_plots: bool = True
    wait_time: float = 0.1
    run_analysis: bool = True
    run_serially: bool = False
    serial_wait_time: float = 10.
    g3_tag: Optional[str] = None
    analysis_kwargs: Optional[dict] = None

    def __post_init__(self):
        if self.biases is None:
            self.biases = np.arange(self.bias_high, self.bias_low - self.bias_step, -self.bias_step)
        self.biases = np.sort(self.biases)[::-1]

        if self.analysis_kwargs is None:
            self.analysis_kwargs = {}


@sdl.set_action()
def take_iv(S, cfg, **kwargs):
    """
    Takes an IV.

    This function requires an accurate bias-group-map, so be sure to run
    ``take_bgmap`` before this.

    Args
    ----
    S : SmurfControl
        pysmurf instance
    cfg : DetConfig
        detconfig instance
    bias_groups : list, int
        List of bias groups to run on. Defaults to all 12.
    overbias_voltage : float
        Voltage to use to overbias detectors
    overbias_wait : float
        Time (sec) to wait at overbiased voltage
    high_current_mode : bool
        If True, will run IV in high-current-mode. This is highly recommended
        to avoid the bias-line filter.
    cool_wait : float
        Amout of time to wait at the first bias step after overbiasing.
    cool_voltage : float
        TES bias voltage to sit at during the cool_wait time while the cryostat
        cools
    biases : np.ndarray, optional
        array of biases to use for IV.
        This should be in units of Low-Current-Mode volts. If you are running
        in high-current-mode this will automatically be adjusted for you!!
    bias_high : float
        Upper limit for biases if not manually set. (to be used in np.arange)
        This should be in units of Low-Current-Mode volts. If you are running
        in high-current-mode this will automatically be adjusted for you!!
    bias_low : float
        Lower limit for biases if not manually set. (to be used in np.arange)
        This should be in units of Low-Current-Mode volts. If you are running
        in high-current-mode this will automatically be adjusted for you!!
    bias_step : float
        Step size for biases if not manually set. (to be used in np.arange)
        This should be in units of Low-Current-Mode volts. If you are running
        in high-current-mode this will automatically be adjusted for you!!
    wait_time : float
        Amount of time to wait at each bias point.
    run_analysis : bool
        If True, will automatically run analysis, save it, and update device
        cfg. (unless otherwise specified in analysis_kwargs)
    show_plots : bool
        If true, will show summary plots
    run_serially : bool
        If True, will run IV sweeps in serial, running each of the specified
        bias groups independently instead of all together.
    serial_wait_time : float
        Time to sleep between serial IV sweeps (sec)
    g3_tag : string (optional)
        If not None, overrides default tag "oper,iv" sent to g3 file.
    analysis_kwargs : dict
        Keyword arguments to pass to analysis

    Returns
    -------
    iva : IVAnalysis
        IVAnalysis object. Will be already analyzed if run_analysis=True.
    """
    if not hasattr(S, 'tune_file'):
        raise AttributeError('No tunefile loaded in current '
                             'pysmurf session. Load active tunefile.')

    kw = deepcopy(cfg.dev.exp['iv_defaults'])
    kw.update(kwargs)
    ivcfg = IVConfig(**kw)

    if ivcfg.bias_groups is None:
        ivcfg.bias_groups = cfg.dev.exp['active_bgs']
    ivcfg.bias_groups = np.atleast_1d(ivcfg.bias_groups).astype(int)

    run_kwargs = asdict(ivcfg)

    start_times = np.zeros((S._n_bias_groups, len(ivcfg.biases)))
    stop_times = np.zeros((S._n_bias_groups, len(ivcfg.biases)))
    sdl.stop_point(S)
    def overbias_and_sweep(bgs, cool_voltage=None):
        """
        Helper function to run IV sweep for a single bg or group of bgs.
        """
        bgs = np.atleast_1d(bgs).astype(int)
        _bias_arr = S.get_tes_bias_bipolar_array()
        _bias_arr[ivcfg.bias_groups] = 0
        S.set_tes_bias_bipolar_array(_bias_arr)
        if ivcfg.overbias_voltage > 0:
            if cool_voltage is None:
                cool_voltage = np.max(ivcfg.biases)
            S.overbias_tes_all(
                bias_groups=bgs, overbias_wait=ivcfg.overbias_wait,
                tes_bias=cool_voltage, cool_wait=ivcfg.cool_wait,
                high_current_mode=ivcfg.high_current_mode,
                overbias_voltage=ivcfg.overbias_voltage
            )

        S.log(f"Starting TES Bias Ramp on bg {bgs}")
        for i, bias in enumerate(ivcfg.biases):
            sdl.stop_point(S)
            S.log(f"Setting bias to {bias:4.3f}", S.LOG_INFO)
            _bias_arr[bgs] = bias
            S.set_tes_bias_bipolar_array(_bias_arr)
            start_times[bgs, i] = time.time()
            time.sleep(ivcfg.wait_time)
            stop_times[bgs, i] = time.time()

    if ivcfg.high_current_mode:
        ivcfg.biases /= S.high_low_current_ratio
    try:
        sid = sdl.stream_g3_on(S, tag=ivcfg.g3_tag, subtype='iv')
        if ivcfg.run_serially:
            for bg in ivcfg.bias_groups:
                overbias_and_sweep(bg, cool_voltage=ivcfg.cool_voltage)
                time.sleep(ivcfg.serial_wait_time)
        else:
            overbias_and_sweep(ivcfg.bias_groups)
    finally:
        sdl.stream_g3_off(S)

        # Turn off biases and streaming on error
        for bg in ivcfg.bias_groups:
            S.set_tes_bias_bipolar(bg, 0)


    iva = IVAnalysis(S, cfg, run_kwargs, sid, start_times, stop_times)

    if ivcfg.run_analysis:
        sdl.stop_point(S)
        _analysis_kwargs = {'save': True, 'update_cfg': True}
        _analysis_kwargs.update(ivcfg.analysis_kwargs)
        analyze_iv(iva, **_analysis_kwargs)

        # Save and publish plots
        is_interactive = plt.isinteractive()
        try:
            if not ivcfg.show_plots:
                plt.ioff()
            fig, ax = plot_Rfracs(iva)
            fname = sdl.make_filename(S, 'iv_rfracs.png', plot=True)
            fig.savefig(fname)
            S.pub.register_file(fname, 'iv', format='png', plot=True)
            if not ivcfg.show_plots:
                plt.close(fig)

            fig, ax = plot_Rn_hist(iva)
            fname = sdl.make_filename(S, 'iv_rns.png', plot=True)
            fig.savefig(fname)
            S.pub.register_file(fname, 'iv', format='png', plot=True)
            if not ivcfg.show_plots:
                plt.close(fig)
        finally:
            if is_interactive:
                plt.ion()

    return iva


def get_plot_text(iva):
    """
    Gets text to add to text-box for IV plots.
    """
    return '\n'.join([
        f"stream_id: {iva.meta['stream_id']}",
        f"sid: {iva.sid}",
        f"path: {os.path.basename(iva.filepath)}",
    ])




"""
Module for miscellaneous functions and classes that are useful in many sodetlib
scripts.
"""
import numpy as np
from scipy import signal
import time
import os
from collections import namedtuple
import sodetlib
from sodetlib import det_config
from sodetlib.constants import *
from sotodlib.tod_ops.fft_ops import calc_psd
from sotodlib.core import AxisManager

if not os.environ.get('NO_PYSMURF', False):
    try:
        import epics
        import pysmurf
        from pysmurf.client.command.cryo_card import cmd_make
    except Exception:
        os.environ['NO_PYSMURF'] = True

StreamSeg = namedtuple("StreamSeg", "times sig mask")

class TermColors:
    HEADER = '\n\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def cprint(msg, style=TermColors.OKBLUE):
    if style is True:
        style = TermColors.OKGREEN
    elif style is False:
        style = TermColors.FAIL
    print(f"{style}{msg}{TermColors.ENDC}")


def make_filename(S, name, ctime=None, plot=False):
    """
    Creates a timestamped filename in the pysmurf outputs or plot directory.
    """
    if ctime is None:
        ctime = S.get_timestamp()

    if plot:
        ddir = S.plot_dir
    else:
        ddir = S.output_dir

    return os.path.join(ddir, f'{ctime}_{name}')


def _encode_data(data):
    """
    Encodes a data object into one that is serializable with
    json so that it can be sent over UDP.
    """
    if isinstance(data, list):
        return [_encode_data(d) for d in data]
    elif isinstance(data, dict):
        return {str(k): _encode_data(d) for k, d in data.items()}
    elif isinstance(data, np.ndarray):
        return data.tolist()
    else:
        return data


def set_session_data(S, key, val):
    """
    Adds data to the OCS Session data object. If this is being run directly
    from an OCS PysmurfController operation, the _ocs_session object will be
    set in the pysmurf instance. If not, publish message so pysmurf monitor can
    relay data to the correct controller agent.
    """
    session = getattr(S, '_ocs_session', None)
    if session is not None:
        session.data[key] = _encode_data(val)
    else:
        S.pub.publish(_encode_data({key: val}), msgtype='session_data')



def pub_ocs_log(S, msg, log=True):
    """
    Passes a string to the OCS pysmurf controller to be logged to be passed
    around the OCS network.
    """
    if log:
        S.log(msg)

    session = getattr(S, '_ocs_session', None)
    if session is not None:
        session.add_message(msg)

    S.pub.publish(msg, msgtype='session_log')


class OCSAbortError(Exception):
    """Error that's thrown when an operation is aborted through OCS""" 


def stop_point(S):
    """
    This function will throw an error if the OCS session status is 'stopping'.
    This allows the pysmurf controller to gracefully tell long-running sodetlib
    functions to stop when possible without killing the agent.
    """
    session = getattr(S, '_ocs_session', None)
    if session is not None:
        if session.status == 'stopping':
            raise OCSAbortError()



def load_bgmap(bands, channels, bgmap_file):
    """
    Loads bias-group assignments and channel polarity for
    set of channels.

    Args
    ----
    bands : np.ndarray
        Array of len <nchans> containing the smurf-band of each channel
    channels : np.ndarray
        Array of len <nchans> containing the smurf-channel of each channel
    bgmap_file : str
        Path to the bgmap file.

    Returns
    --------
    bgmap : np.ndarray
        Array of lenght <nchans> containing bias-group assignments of each
        channel. Unassigned channels contain -1.
    polarity : np.ndarray
        Array of lenght <nchans> containing the polarity of each channel,
        or whether the squid current steps up or down with bias-current.
    """
    bgmap_full = np.load(bgmap_file, allow_pickle=True).item()
    idxs = map_band_chans(
        bands, channels, bgmap_full['bands'], bgmap_full['channels']
    )
    bgmap = bgmap_full['bgmap'][idxs]
    polarity = bgmap_full['polarity'][idxs]
    bgmap[idxs == -1] = -1

    return bgmap, polarity


def map_band_chans(b1, c1, b2, c2, chans_per_band=512):
    """
    Returns an index mapping of length nchans1 from one set of bands and
    channels to another. Note that unmapped indices are returned as -1, so this
    must be handled before (or after) indexing or else you'll get weird
    results.

    Args
    ----
        b1 (np.ndarray):
            Array of length nchans1 containing the smurf band of each channel
        c1 (np.ndarray):
            Array of length nchans1 containing the smurf channel of each channel
        b2 (np.ndarray):
            Array of length nchans2 containing the smurf band of each channel
        c2 (np.ndarray):
            Array of length nchans2 containing the smurf channel of each channel
        chans_per_band (int):
            Lets just hope this never changes.
    """
    acs1 = b1 * chans_per_band + c1
    acs2 = b2 * chans_per_band + c2

    mapping = np.full_like(acs1, -1, dtype=int)
    for i, ac in enumerate(acs1):
        x = np.where(acs2 == ac)[0]
        if len(x > 0):
            mapping[i] = x[0]

    return mapping


def get_metadata(S, cfg):
    """
    Gets collection of metadata from smurf and detconfig instance that's
    useful for pretty much everything. This should be included in all output
    data files created by sodetlib.
    """
    return {
        'tunefile': getattr(S, 'tune_file', None),
        'high_low_current_ratio': S.high_low_current_ratio,
        'R_sh': S.R_sh,
        'pA_per_phi0': S.pA_per_phi0,
        'rtm_bit_to_volt': S._rtm_slow_dac_bit_to_volt,
        'bias_line_resistance': S.bias_line_resistance,
        'chans_per_band': S.get_number_channels(),
        'high_current_mode': get_current_mode_array(S),
        'timestamp': time.time(),
        'stream_id': cfg.stream_id,
        'action': S.pub._action,
        'action_timestamp': S.pub._action_ts,
        'bgmap_file': cfg.dev.exp.get('bgmap_file'),
        'iv_file': cfg.dev.exp.get('iv_file'),
        'v_bias': S.get_tes_bias_bipolar_array(),
        'pysmurf_client_version': pysmurf.__version__,
        'rogue_version': S._caget(f'{S.epics_root}:AMCc:RogueVersion'),
        'smurf_core_version': S._caget(f'{S.epics_root}:AMCc:SmurfApplication:SmurfVersion'),
        'sodetlib_version': sodetlib.__version__,
        'fpga_git_hash': S.get_fpga_git_hash_short(),
        'cryocard_fw_version': S.C.get_fw_version(),
        'crate_id': cfg.sys['crate_id'],
        'slot': cfg.slot,
    }


def validate_and_save(fname, data, S=None, cfg=None, register=True,
                      make_path=True):
    """
    Does some basic data validation for sodetlib data files to make sure
    they are normalized based on the following  rules:
        1. The 'meta' key exists storing metadata from smurf and cfg instances
           such as the stream-id, pysmurf constants like high-low-current-ratio
           etc., action and timestamp and more. See the get_metadata function
           definition for more.
        2. 'bands', and 'channels' arrays are defined, specifying the band/
           channel combinations for data taking / analysis products.
        3. 'sid' field exists to contain one or more session-ids for the
           analysis.

    Args
    ----
        fname (str):
            Filename or full-path of datafile. If ``make_path`` is set to True,
            this will be turned into a file-path using the ``make_filename``
            function. If not, this will be treated as the save-file-path
        data (dict):
            Data to be written to disk. This should contain at least the keys
            ``bands``, ``channels``, ``sid``, along with any other data
            products. If the key ``meta`` does not already exist, metadata
            will be populated using the smurf and det-config instances.
        S (SmurfControl):
            Pysmurf instance. This must be set if registering data file,
            ``meta`` is not yet set, or ``make_path`` is True.
        cfg (DetConfig):
            det-config instance. This must be set if registering data file,
            ``meta`` is not yet set, or ``make_path`` is True.
        register (bool):
            If True, will register the file with the pysmurf publisher
        make_path (bool):
            If true, will create a path by passing fname to the function
            make_filename (putting it in the pysmurf output directory)
    """
    if register or make_path or 'meta' not in data:
        if S is None or cfg is None:
            raise ValueError("Pysmurf and cfg instances must be set")

    _data = data.copy()
    if 'meta' not in data:
        meta = get_metadata(S, cfg)
        _data['meta'] = meta

    for k in ['channels', 'bands', 'sid']:
        if k not in data:
            raise ValueError(f"Key '{k}' is required in data")

    if make_path and S is not None:
        path = make_filename(S, fname)
    else:
        path = fname

    np.save(path, _data, allow_pickle=True)
    if S is not None and register:
        S.pub.register_file(path, 'sodetlib_data', format='npy')
    return path


def get_tracking_kwargs(S, cfg, band, kwargs=None):
    band_cfg = cfg.dev.bands[band]
    tk = {
        'reset_rate_khz': cfg.dev.exp['flux_ramp_rate_khz'],
        'lms_freq_hz': band_cfg['lms_freq_hz'],
        'lms_gain': band_cfg['lms_gain'],
        'fraction_full_scale': band_cfg['frac_pp'],
        'make_plot': True,
        'show_plot': True,
        'channel': [],
        'nsamp': 2**18,
        'return_data': True,
        'feedback_start_frac': cfg.dev.exp['feedback_start_frac'],
        'feedback_end_frac': cfg.dev.exp['feedback_end_frac'],
        'feedback_gain': band_cfg['feedback_gain'],
    }
    if kwargs is not None:
        tk.update(kwargs)
    return tk

def get_psd(S, times, phases, detrend='constant', nperseg=2**12, fs=None):
    """
    Returns PSD for all channels.
    Args:
        S:
            pysmurf.SmurfControl object
        times: np.ndarray
            timestamps (in ns)
        phases: np.ndarray
            Array of phases
        detrend: str
            Detrend argument to pass to signal.welch
        nperseg: int
            nperseg arg for signal.welch
        fs: float
            sample frequency for signal.welch. If None will calculate using the
            timestamp array.
    Returns:
        f: np.ndarray
            Frequencies
        Pxx: np.ndarray
            PSD in pA/sqrt(Hz)
    """
    if fs is None:
        fs = 1/np.diff(times/1e9).mean()
    current = phases * S.pA_per_phi0 / (2 * np.pi)
    f, Pxx = signal.welch(current, detrend=detrend, nperseg=nperseg, fs=fs)
    Pxx = np.sqrt(Pxx)
    return f, Pxx

def get_asd(am, pA_per_phi0=9e6, **psd_kwargs):
    """
    Returns ASD (sqrt(PSD)) for all channels.

    Args
    ----
    am: AxisManager
        timestamps (in ns)
    pA_per_phi0: float
        Conversion from phi_0 to pA, defaults to 9e6 set by the mux chip
        mutual inductance between TES to SQUID.
    psd_kwargs: dictionary
        keyword arguments taken by scipy.welch function.

    Returns:
        f: np.ndarray
            Frequencies
        Axx: np.ndarray
            ASD in pA/sqrt(Hz)
    """
    f, Pxx = calc_psd(am, **psd_kwargs)
    Axx = np.sqrt(Pxx)*pA_per_phi0/(2 * np.pi)
    return f, Axx


class SectionTimer:
    def __init__(self):
        self.sections = []
        self.start_time = None
        self.stop_time = None

    def start_section(self, name):
        if self.start_time is None:
            self.start_time = time.time()
        self.sections.append((time.time(), name))

    def stop(self):
        self.stop_time = time.time()
        self.sections.append((time.time(), 'STOP'))

    def reset(self):
        self.sections = []
        self.start_time = None
        self.stop_time = None

    def summary(self):
        out = "="*80 + '\nTiming Summary\n' + '-'*80 + '\n'
        out += f"Total time: {self.stop_time - self.start_time} sec\n"
        out += 'name\tdur\tstart\n' + '='*80 + '\n'

        name_len = max([len(name) for t, name in self.sections])

        for i in range(len(self.sections) - 1):
            t, name = self.sections[i]
            dur = self.sections[i+1][0] - t
            out += f'{name:{name_len}s}\t{dur:.2f}\t{t:.0f}\n'

        return out


def dev_cfg_from_pysmurf(S, save_file=None, clobber=True):
    """
    Creates a populated device cfg object from a fully tuned pysmurf instance.
    If a save-file is specifed, the device config file will be written there.
    By default this will not save the device config to a file!! If you want
    overwrite the currently used device cfg, you can run::

        dev_cfg_from_pysmurf(S, save_file=cfg.dev_file, clobber=True)

    Args
    ----
    S : SmurfControl object
        The pysmurf instance should be in a state where a tunefile is loaded,
        attenuations, biases, and any other parameter are already set
        correctly.
    save_file : path
        Path to save-file location. Remember that if you are running in a
        docker container, you have to give the path as it is inside the
        container. For example, the OCS_CONFIG_DIR is mapped to /config inside
        the docker.
    clobber : bool
        If true, will overwrite the save_file if one already exists at that
        location.
    """
    dev = det_config.DeviceConfig()

    # Experiment setup
    amp_biases = S.get_amplifier_biases()
    if hasattr(S, 'tune_file'):
        tunefile = S.tune_file
    else:
        cprint("No tunefile is loaded! Loading tunefile=None", False)
        tunefile = None
    dev.exp.update({
        'amp_50k_Id': amp_biases['50K_Id'],
        'amp_50k_Vg': amp_biases['50K_Vg'],
        'amp_hemt_Id': amp_biases['hemt_Id'],
        'amp_hemt_Vg': amp_biases['hemt_Vg'],
        'tunefile': tunefile,
        'bias_line_resistance': S._bias_line_resistance,
        'high_low_current_ratio': S._high_low_current_ratio,
        'pA_per_phi0': S._pA_per_phi0,
    })

    # Right now not getting any bias group info
    for band in S._bands:
        tone_powers = S.get_amplitude_scale_array(band)[S.which_on(band)]
        if len(tone_powers) == 0:
            drive = S._amplitude_scale[band]
            cprint(f"No channels are on in band {band}. Setting drive to "
                   f"pysmurf-cfg value: {drive}", style=TermColors.WARNING)
        else:
            drives, counts = np.unique(tone_powers, return_counts=True)
            drive = drives[np.nanargmax(counts)]
            if len(drives) > 1:
                print(f"Multiple drive powers exist for band {band} ({drives})!")
                print(f"Using most common power: {drive}")

        feedback_start_frac = S._feedback_to_feedback_frac(band, S.get_feedback_start(band))
        feedback_end_frac = S._feedback_to_feedback_frac(band, S.get_feedback_end(band))

        flux_ramp_rate_khz = S.get_flux_ramp_freq()
        lms_freq_hz = S.get_lms_freq_hz(band)
        nphi0 = np.round(lms_freq_hz / flux_ramp_rate_khz / 1e3)

        dev.bands[band].update({
            'uc_att': S.get_att_uc(band),
            'dc_att': S.get_att_dc(band),
            'drive': drive,
            'feedback_start_frac': feedback_start_frac,
            'feedback_end_frac': feedback_end_frac,
            'lms_gain': S.get_lms_gain(band),
            'frac_pp': S.get_fraction_full_scale(),
            'flux_ramp_rate_khz': flux_ramp_rate_khz,
            'lms_freq_hz': lms_freq_hz,
            'nphi0': nphi0
        })

    if save_file is not None:
        if clobber and os.path.exists(save_file):
            print(f"Rewriting existing file: {save_file}")
        dev.dump(save_file, clobber=clobber)
    return dev


def get_wls_from_am(am, nperseg=2**16, fmin=10., fmax=20., pA_per_phi0=9e6):
    """
    Gets white-noise levels for each channel from the axis manager returned
    by smurf_ops.load_session.

    Args
    ----
    am : AxisManager
        Smurf data returned by so.load_session or the G3tSmurf class
    nperseg : int
        nperseg to be passed to welch
    fmin : float
        Min frequency to use for white noise mask
    fmax : float
        Max freq to use for white noise mask
    pA_per_phi0 : float
        S.pA_per_phi0 unit conversion. This will eventually make its way
        into the axis manager, but for now I'm just hardcoding this here
        as a keyword argument until we get there.

    Returns
    --------
    wls : array of floats
        Array of the white-noise level for each channel, indexed by readout-
        channel number
    band_medians : array of floats
        Array of the median white noise level for each band.
    """
    fsamp = 1./np.median(np.diff(am.timestamps))
    fs, pxx = signal.welch(am.signal * pA_per_phi0 / (2*np.pi),
                           fs=fsamp, nperseg=nperseg)
    pxx = np.sqrt(pxx)
    fmask = (fmin < fs) & (fs < fmax)
    wls = np.median(pxx[:, fmask], axis=1)
    band_medians = np.zeros(8)
    for i in range(8):
        m = am.ch_info.band == i
        band_medians[i] = np.median(wls[m])
    return wls, band_medians


# useful functions for analysis etc.
def invert_mask(mask):
    """
    Converts a readout mask from (band, chan)->rchan form to rchan->abs_chan
    form.
    """
    bands, chans = np.where(mask != -1)
    maskinv = np.zeros_like(bands, dtype=np.int16)
    for b, c in zip(bands, chans):
        maskinv[mask[b, c]] = b * CHANS_PER_BAND + c
    return maskinv


def get_r2(sig, sig_hat):
    """ Gets r-squared value for a signal"""
    sst = np.sum((sig - sig.mean())**2)
    sse = np.sum((sig - sig_hat)**2)
    r2 = 1 - sse / sst
    if r2 < 0:
        return 0
    return r2


class _Register:
    def __init__(self, S, addr):
        self.S = S
        self.addr = S.epics_root + ":" + addr

    def get(self, **kw):
        return self.S._caget(self.addr, **kw)

    def set(self, val, **kw):
        self.S._caput(self.addr, val, **kw)

class Registers:
    """
    Utility class for storing, getting, and setting SO-rogue registers even if
    they are not in the standard rogue tree, or settable by existing pysmurf
    get/set functions
    """
    _root = 'AMCc:'
    _processor = _root + "SmurfProcessor:"
    _sostream = _processor + "SOStream:"
    _sofilewriter = _sostream + 'SOFileWriter:'
    _source_root = _root + 'StreamDataSource:'

    _registers = {
        'pysmurf_action': _sostream + 'pysmurf_action',
        'pysmurf_action_timestamp': _sostream + "pysmurf_action_timestamp",
        'stream_tag': _sostream + "stream_tag",
        'open_g3stream': _sostream + "open_g3stream",
        'g3_session_id': _sofilewriter + "session_id",
        'g3_filepath': _sofilewriter + "filepath",
        'debug_data': _sostream + "DebugData",
        'debug_meta': _sostream + "DebugMeta",
        'debug_builder': _sostream + "DebugBuilder",
        'flac_level': _sostream + "FlacLevel",
        'source_enable': _source_root + 'SourceEnable',
        'enable_compression': _sostream + 'EnableCompression',
        'agg_time': _sostream + 'AggTime',
    }

    def __init__(self, S):
        self.S = S
        for name, reg in self._registers.items():
            setattr(self, name, _Register(S, reg))


def get_current_mode_array(S):
    """
    Gets high-current-mode relay status for all bias groups
    """
    relay = S.get_cryo_card_relays()
    relay = S.get_cryo_card_relays()  # querey twice to ensure update
    bgs = np.arange(S._n_bias_groups)
    hcms = np.zeros_like(bgs, dtype=bool)
    for i, bg in enumerate(bgs):
        r = np.ravel(S._pic_to_bias_group[np.where(
            S._pic_to_bias_group[:, 1] == bg)])[0]
        hcms[i] = (relay >> r) & 1
    return hcms


def set_current_mode(S, bgs, mode, const_current=True):
    """
    Sets one or more bias lines to high current mode. If const_current is True,
    will also update the DC bias voltages to try and preserve current based on
    the set high_low_current_ratio. We need this function to replace the
    existing pysmurf function so we can set both PV's in a single epics call,
    minimizing heating on the cryostat.

    This function will attempt to check the existing rogue relay state to
    determine if the voltages need to be updated or not. This may not work all
    of the time since there is no relay readback for the high-current-mode
    relays.  That means this will set the voltages incorrectly if there is
    somehow an inconsistency between the rogue relay state and the real relay
    state.

    Args
    ----
        S : SmurfControl
            Pysmurf control instance
        bgs : (int, list)
            Bias groups to switch to high-current-mode
        mode : int
            1 for high-current, 0 for low-current
        const_current : bool
            If true, will adjust voltage values simultaneously based on
            S.high_low_current_ratio
    """

    bgs = np.atleast_1d(bgs).astype(int)

    # DO this twice bc pysmurf does for some reason
    old_relay = S.get_cryo_card_relays()
    old_relay = S.get_cryo_card_relays()
    new_relay = np.copy(old_relay)

    old_dac_volt_arr = S.get_rtm_slow_dac_volt_array()
    new_dac_volt_arr = np.copy(old_dac_volt_arr)

    for bg in bgs:
        # Gets relay bit index for bg
        idx = np.where(S._pic_to_bias_group[:, 1] == bg)[0][0]
        # Bit in relay mask corresponding to this bg's high-current
        r = S._pic_to_bias_group[idx, 0]

        # Index of bg's DAC pair
        pair_idx = np.where(S._bias_group_to_pair[:, 0] == bg)[0][0]
        pos_idx = S._bias_group_to_pair[pair_idx, 1] - 1
        neg_idx = S._bias_group_to_pair[pair_idx, 2] - 1
        if mode:
            # sets relay bit to 1
            new_relay = new_relay | (1 << r)

            # if const_current and the old_relay bit was zero, divide bias
            # voltage by high_low_ratio
            if const_current and not (old_relay >> r) & 1:
                new_dac_volt_arr[pos_idx] /= S.high_low_current_ratio
                new_dac_volt_arr[neg_idx] /= S.high_low_current_ratio
        else:
            # Sets relay bit to 0
            new_relay = new_relay & ~(1 << r)

            # if const_current and the old_relay bit was one, mult bias
            # voltage by high_low_ratio
            if const_current and (old_relay >> r) & 1:
                new_dac_volt_arr[pos_idx] *= S.high_low_current_ratio
                new_dac_volt_arr[neg_idx] *= S.high_low_current_ratio

    relay_data = cmd_make(0, S.C.relay_address, new_relay)

    # Sets DAC data values, clipping the data array to make sure they contain
    # the correct no. of bits
    dac_data = (new_dac_volt_arr / S._rtm_slow_dac_bit_to_volt).astype(int)
    nbits = S._rtm_slow_dac_nbits
    dac_data = np.clip(dac_data, -2**(nbits-1), 2**(nbits-1)-1)

    dac_data_reg = S.rtm_spi_max_root + S._rtm_slow_dac_data_array_reg


    if isinstance(S.C.writepv, str):
        cryocard_writepv = S.C.writepv
    else:
        cryocard_writepv = S.C.writepv.pvname

    # It takes longer for DC voltages to settle than it does to toggle the
    # high-current relay, so we can set them at the same time when switchign
    # to hcm, but when switching to lcm we need a sleep statement to prevent
    # dets from latching.
    if mode:
        epics.caput_many([cryocard_writepv, dac_data_reg], [relay_data, dac_data],
                         wait=True)
    else:
        S._caput(dac_data_reg, dac_data)
        time.sleep(0.04)
        S._caput(cryocard_writepv, relay_data)

    time.sleep(0.1)  # Just to be safe


def save_fig(S, fig, name, tag=''):
    """
    Saves figure to pysmurf plots directory and publishes result.

    Args
    -----
    S : SmurfControl
        Pysmurf instantc
    fig : Figure
        Matplotlib figure to save
    name : Figure name
        This will be passed to ``make_filename`` so it is timestamped
        and placed in the pysmurf plots directory.
    """
    fname = make_filename(S, name, plot=True)
    fig.savefig(fname)
    S.pub.register_file(fname, tag, plot=True)
    return fname


################################################################################
# AxisManager Utility Functions
################################################################################

class RestrictionException(Exception):
    """Exception for when cannot restrict AxisManger properly"""

def restrict_to_times(am, t0, t1, in_place=False):
    """
    Restricts axis manager to a time range (t0, t1)
    """
    m = (t0 < am.timestamps) & (am.timestamps < t1)
    if not m.any():
        raise RestrictionException
    i0, i1 = np.where(m)[0][[0, -1]] + am.samps.offset
    return am.restrict('samps', (i0, i1), in_place=in_place)

def dict_to_am(d, skip_bad_types=False):
    """
    Attempts to convert a dictionary into an axis manager. This can be used on
    dicts containing basic types such as (str, int, float) along with numpy
    arrays. The AxisManager will not have any structure such as "axes", but
    this is useful if you want to nest semi-arbitrary data such as the "meta"
    dict into an axis manager.

    Args
    -----
    d : dict
        Dict to convert ot axismanager
    skip_bad_types : bool
        If True, will skip any value that is not a str, int, float, or
        np.ndarray. If False, this will raise an error for invalid types.
    """
    allowed_types = (str, int, float, np.ndarray)
    am = AxisManager()
    for k, v in d.items():
        if isinstance(v, dict):
            am.wrap(k, dict_to_am(v))
        if isinstance(v, allowed_types):
            am.wrap(k, v)
        elif not skip_bad_types:
            raise ValueError(
                f"Key {k} is of type {type(v)} which cannot be wrapped by an "
                 "axismanager")
    return am

def remap_dets(src, dst, load_axes=False, idxmap=None):
    """
    This function takes in two axis-managers, and returns a copy of the 2nd
    which is identical except all fields aligned with 'dets' are remapped so they
    match up with the 'dets' field of the first array. This is very useful if you
    want to compare data across tunes, or across cryostats. An additional field
    "unmapped" of shape ``(dets, )`` will be added containing mask for which
    channels were unmapped by the idxmap.

    Args
    ----
    src : AxisManager
        "source" data. A copy of this data will be returned, with the 'dets'
        axis remapped to match those of the dest AxisManager
    dst : AxisManager
        "destination" data. The 'dets' axis will be taken from this AxisManager.
    load_axes : bool
        If True, will add ``src`` axes into the copy, and copy the axis
        assignments. Note that if you are trying to nest this in another
        axismanager, this may cause trouble if 'src', and 'dst' have 
        axes with the same name but different values. If this is false,
        only the 'dets' assignment will be used, allowing for safe nesting.
    idxmap : optional, np.ndarry
        If an idxmap is passed, that will be used to remap the dets axis.
        If none is passed, an idxmap will be created to match up bands and
        channels of the src and dst. This might be useful if you want to
        map across cooldowns based on a detmap.
    """
    if idxmap is None:
        idxmap = map_band_chans(
            dst.bands, dst.channels,
            src.bands, src.channels,
        )

    axes = [dst.dets]
    if load_axes:
        axes.extend([v for k, v in src._axes.items() if k != 'dets'])

    am_new = AxisManager(*axes)
    for k, axes in src._assignments.items():
        f = src._fields[k]
        if isinstance(f, AxisManager):
            am_new.wrap(k, remap_dets(f, dst, load_axes=load_axes, idxmap=idxmap))
        elif isinstance(f, np.ndarray):
            assignments = []
            if load_axes:
                assignments.extend([
                    (i, n) for i, n in enumerate(axes) 
                    if n is not None and n!='dets'
                ])
            if 'dets' in axes:
                i = axes.index('dets')
                f = np.rollaxis(np.rollaxis(f, i)[idxmap], -i)
            am_new.wrap(k, f, assignments)
        else:  # Literals
            am_new.wrap(k, f)

    am_new.wrap('unmapped', idxmap == -1, [(0, 'dets')])
    return am_new


def overbias_dets(S, cfg, bias_groups=None, biases=None, cool_wait=None,
                  high_current_mode=False):
    """
    Overbiases detectors based on parameters in the device cfg.

    Args
    ------
    S : SmurfControl
        Pysmurf control instance
    cfg : DetConfig
        DetConfig instance
    bias_groups : list
        List of bias groups to overbias
    biases: list
        Array of final bias values to set after overbiasing. This must
        be an array of length 12 where biases[bg] is the bias of bias group bg.
        If this is not set, the ``cool_voltage`` set in the device cfg file
        will be used for each bias group.
    cool_wait : float
        If set, time to wait after overbiasing dets. Defaults to no wait time.
    high_current_mode : bool, List[bool]
        Whether the bias groups should be left in high-current-mode. If a list is
        passed, it should be an array of len 12 where high_current_mode[bg] is
        the desired mode for bias line ``bg``.
    """
    if bias_groups is None:
        bias_groups = cfg.dev.exp['active_bgs']

    if isinstance(high_current_mode, (bool, int)):
        high_current_mode = [high_current_mode for _ in range(12)]
    high_current_mode = np.atleast_1d(high_current_mode)

    S.log("Overbiasing Detectors")
    set_current_mode(S, bias_groups, 1, const_current=False)
    for bg in bias_groups:
        S.set_tes_bias_bipolar(bg, cfg.dev.bias_groups[bg]['overbias_voltage'])

    wait_time = cfg.dev.exp['overbias_wait']
    S.log(f"Waiting at ob volt for {wait_time} sec")
    time.sleep(wait_time)

    for bg in bias_groups:
        if biases is not None:
            bias = biases[bg]
        else:
            bias = cfg.dev.bias_groups[bg]['cool_voltage']
        S.set_tes_bias_bipolar(bg, bias)

    lcm_bgs = np.where(~high_current_mode)[0]
    set_current_mode(S, lcm_bgs, 0, const_current=False)

    if cool_wait is not None:
        time.sleep(cool_wait)

    return S.get_tes_bias_bipolar_array()



import numpy as np
from tqdm.auto import tqdm, trange
import time
import numpy as np
import sodetlib as sdl
from sodetlib.constants import *
from scipy.signal import welch, hilbert
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from sotodlib.tod_ops.filters import gaussian_filter, fourier_filter 
from sotodlib.core import IndexAxis, AxisManager, OffsetAxis, LabelAxis


def new_ci_dset(S, cfg, bands, chans, freqs, run_kwargs=None, ob_path=None,
                sc_path=None):
    """
    Creates a new CIData AxisManager. If ob_path and sc_path are set, they
    will be loaded with a remapped "dets" axis.
    """
    ndets = len(chans)
    nsteps = len(freqs)

    ds = AxisManager(
        LabelAxis('dets', vals=[f"r{x:0>4}" for x in range(ndets)]),
        IndexAxis('steps', count=nsteps),
        IndexAxis('biaslines', count=NBGS),
    )
    ds.wrap('meta', sdl.dict_to_am(sdl.get_metadata(S, cfg), skip_bad_types=True))
    ds.meta.wrap('g3_dir', cfg.sys['g3_dir'])
    if run_kwargs is not None:
        ds.wrap('run_kwargs', sdl.dict_to_am(run_kwargs))

    ds.wrap('bands', bands, [(0, 'dets')])
    ds.wrap('channels', chans, [(0, 'dets')])
    ds.wrap('freqs', freqs, [(0, 'steps')])

    ds.wrap_new('start_times', ('biaslines', 'steps'))
    ds.wrap_new('stop_times', ('biaslines', 'steps'))
    ds.wrap_new('sids', ('biaslines',), dtype=int)

    # Load bgmap stuff
    bgmap, polarity = sdl.load_bgmap(
        bands, chans, cfg.dev.exp['bgmap_file']
    )
    ds.wrap('bgmap', bgmap, [(0, 'dets')])
    ds.wrap('polarity', polarity, [(0, 'dets')])


    ob_path = cfg.dev.exp.get('complex_impedance_ob_path')
    sc_path = cfg.dev.exp.get('complex_impedance_sc_path')
    if ds.run_kwargs.state == 'transition':
        if ob_path is None:
            raise ValueError("No ob CI path found in the device cfg")
        if sc_path is None:
            raise ValueError("No sc CI path found in the device cfg")

        ob = sdl.remap_dets(AxisManager.load(ob_path), ds, load_axes=False)
        ds.wrap('ob', ob)

        sc = sdl.remap_dets(AxisManager.load(sc_path), ds, load_axes=False)
        ds.wrap('sc', sc)

    return ds

def A_per_bit(ds):
    return 2 * ds.meta['rtm_bit_to_volt']          \
        / ds.meta['bias_line_resistance']          \
        * ds.meta['high_low_current_ratio']

def load_tod(ds, bg, arc=None):
    """
    Loads a TOD for a biasgroup given a CI dset.

    Args
    -----
    ds : AxisManager
        CI Dataset
    bg : int
        Bias group to load data for
    arc : G3tSmurf, optional
        If a G3tSmurf archive is passed this will be used to load data.
    """
    if arc is not None:
        start = ds.start_times[bg, 0]
        stop = ds.stop_times[bg, -1]
        seg = arc.load_data(start, stop, show_pb=False)
    else:
        sid = ds.sids[bg]
        seg = sdl.load_session(ds.meta.stream_id, sid, base_dir=ds.meta.g3_dir)
    return seg

################################################################################
# CI Analysis
################################################################################
def analyze_seg(ds, tod, bg, i):
    """
    Analyze segment of CI data. The main goal of this is to calculate
    the Ites phasor, containing the amplitude (A) and phase (relative
    tot he commanded) of the TES response to an incoming sine wave.
    This performs the following steps:
     1. Restrict full TOD to a single excitation frequency. This will put
        everything units of A, correct for channel polarity. This will also
        correct timestamps based on the FrameCounter.
     2. Takes PSD of the bias data to obtain the reference freq. Filters signal
        using gaussian filter around reference freq.
     3. Use lock-in amplification with bias as reference to extract
        amplitude and phase of the filtered signal with respect to the
        commanded bias.

    Args
    -----
    ds : AxisManager
        CI Dataset
    tod : AxisManger
        axis-manager containing tod for a given bias group
    bg : int
        Bias group that is being analyzed
    i : int
        Freq index. 0 will analyze the first freq segment taken.

    Returns
    ---------
    am : AxisManager
        Returns an souped-up tod axis-manager corresponding to this freq with
        the following fields:
         - timestmaps, biases, signal, ch_info. Standard tod axismanager stuff
           but with units converted into A, timestamps fixed, and offsets
           subtracted out.
         - sample_rate, cmd_freq: Floats with the sample rate and commanded
           freq
         - filt_sig: Filtered signal (using gaussian filter around commanded
           freq)
         - lockin_x, lockin_y: Lockin x and y signals, used to calc amp and phase
           across the tod
         - Ites: Phasor for Ites. Amplitude is the amp of the sine wave response,
           and angle is the phase relative to the commanded bias.
    """
    t0, t1 = ds.start_times[bg, i], ds.stop_times[bg, i]
    am = sdl.restrict_to_times(tod, t0, t1, in_place=False)

    sample_rate = 1./np.median(np.diff(am.timestamps))

    # Convert everything to A
    am.signal = am.signal * ds.meta['pA_per_phi0'] / (2*np.pi) * 1e-12

    # Index mapping from am readout channel to sweep chan index.
    chan_idxs = sdl.map_band_chans(
        am.ch_info.band, am.ch_info.channel,
        ds.bands, ds.channels
    )
    am.signal *= ds.polarity[chan_idxs, None]

    am.biases = am.biases * A_per_bit(ds)

    # Remove offset from signal
    am.signal -= np.mean(am.signal, axis=1)[:, None]
    am.biases -= np.mean(am.biases, axis=1)[:, None]

    # Fix up timestamps based on frame-counter
    t0, t1 = am.timestamps[0], am.timestamps[-1]
    fc = am.primary["FrameCounter"]
    fc = fc - fc[0]
    ts = t0 + fc/fc[-1] * (t1 - t0)
    am.timestamps = ts
    am.wrap('sample_rate', sample_rate)
    am.wrap('cmd_freq', ds.freqs[i])
    
    # Get psds cause we'll want that
    nsamp = len(am.signal[0])
    fxx, bias_pxx = welch(am.biases[bg], fs=sample_rate, nperseg=nsamp)

    # Gaussian filter around peak freq of bias asd.
    f0, f1 = .9 * am.cmd_freq, 1.1*am.cmd_freq
    m = (f0 < fxx) & (fxx < f1)

    idx = np.argmax(bias_pxx[m])
    f = fxx[m][idx]

    filt = gaussian_filter(f, f_sigma=f / 5)
    filt_sig = fourier_filter(am, filt)
    am.wrap('filt_sig', filt_sig, [(0, 'dets'), (1, 'samps')])

    # Lock in amplification!
    # To get ref + ref offset by 90-deg, take Hilbert transform, and then 
    # real part gives you ref and imag part is offset by 90 deg.
    sig = filt_sig
    ref = hilbert(am.biases[bg] / np.max(am.biases[bg]))
    X = sig * ref.real
    Y = sig * ref.imag
    # We're averaging over enough periods where we don't really need to
    # restrict to an int number of periods...
    xmean = np.mean(X, axis=1)
    ymean = np.mean(Y, axis=1)
    phase = -np.arctan2(ymean, xmean)
    amp = 2*np.sqrt(xmean**2 + ymean**2)
    Ites = amp * np.exp(1.0j * phase)
    am.wrap('lockin_x', X)
    am.wrap('lockin_y', Y)
    am.wrap('Ites', Ites)
    return am

def analyze_tods(ds, bgs=None, tod=None, arc=None, show_pb=True):
    """
    Analyzes TODS for a CIData set. This will add the following fields to the
    dataset:
      - Ites(dets, steps): Ites phasor for each detector / frequency
        combination. Amplitude is the amp of the current response (A), and angle
        is the phase relative to commanded bias.
      - Ibias(biaslines): Amplitude (A) of the sinewave used for each
        biasline.
      - Ibias_dc(biaslines): DC bias current (A) for each biasline
      - res_freqs(dets): Resonance frequency of each channel detectors.
    """
    if bgs is None:
        bgs = ds.run_kwargs.bgs
    bgs = np.atleast_1d(bgs)

    # Delete temp fields if they exist
    for f in ['_Ites', '_Ibias', '_Ibias_dc', '_res_freqs']:
        if f in ds._fields:
            ds.move(f, None)

    nsteps = len(ds.freqs)
    ds.wrap_new('_Ites', ('dets', nsteps), cls=np.full,
                fill_value=np.nan, dtype=np.complex128)
    ds.wrap_new('_Ibias', (NBGS,), cls=np.full, fill_value=np.nan)
    ds.wrap_new('_Ibias_dc', (NBGS,), cls=np.full, fill_value=np.nan)
    ds.wrap_new('_res_freqs', ('dets',), cls=np.full, fill_value=np.nan)

    ntot = len(bgs) * len(ds.freqs)
    pb = tqdm(total=ntot, disable=(not show_pb))
    for bg in bgs:
        if tod is None:
            pb.set_description(f"Loading tod for bg {bg}")
            _tod = load_tod(ds, bg, arc=arc)
        else:
            _tod = tod
        chmap = sdl.map_band_chans(
            _tod.ch_info.band, _tod.ch_info.channel,
            ds.bands, ds.channels
        )

        pb.set_description(f"Analyzing segments for bg {bg}")
        for i in range(len(ds.freqs)):
            try:
                seg = analyze_seg(ds, _tod, bg, i)
            except sdl.RestrictionException:
                # Means there's no data at the specified time
                pb.update()
                continue
            if i == 0:
                ds._Ibias_dc[bg] = np.mean(seg.biases[bg])
                ds._Ibias[bg] = 0.5 * np.ptp(seg.biases[bg])
                ds._res_freqs[chmap] = seg.ch_info.res_frequency

            ds._Ites[chmap, i] = seg.Ites
            pb.update()
        del _tod

    for f in ['_Ites', '_Ibias', '_Ibias_dc', '_res_freqs']:
        ds.move(f, f[1:])

    return ds

def get_ztes(ds):
    """
    Calculates Ztes for in-transition CIData. Adds the following fields to the
    CI dataset:
      - Rn (dets): Normal resistances based off of low-f overbiased data points.
      - Rtes (dets): TES Resistance, based off of low-f in-transition segment
      - Vth (dets): Thevenin equiv voltage (V)
      - Zeq (dets): Equiv impedance
      - Ztes (dets): TES complex impedance
    """
    ob, sc = ds.ob, ds.sc

    fields = ['_Rn', '_Rtes', '_Vth', '_Zeq', '_Ztes']
    for f in fields:
        if f in ds._fields:
            ds.move(f, None)
 
    ds.wrap_new('_Rn', ('dets',))
    ds.wrap_new('_Rtes', ('dets',))
    ds.wrap_new('_Vth', ('dets',))
    ds.wrap_new('_Zeq', ('dets',))
    ds.wrap_new('_Ztes', ('dets',))
   
    # Calculates Rn
    Ib_ob = ob.Ibias[ob.bgmap][:, None]
    ds._Rn = ob.meta.R_sh * (np.abs(Ib_ob / ob.Ites) - 1)[:, 0]

    # Calculate Rtes for in-transition dets
    Ib = ds.Ibias_dc[ds.bgmap]

    dIrat = np.real(ds.Ites[:, 0]) / np.abs(ds.Ibias[ds.bgmap])
    I0 = Ib * dIrat / (2 * dIrat - 1)
    Pj = I0 * ds.meta.R_sh * (Ib - I0)
    ds._Rtes = np.abs(Pj / I0**2)

    Ites_ob = np.zeros_like(ds.Ites)
    Ites_sc = np.zeros_like(ds.Ites)
    for rc in range(len(ob.channels)):
        Ites_ob[rc, :] = np.interp(ds.freqs, ob.freqs, ob.Ites[rc])
        Ites_sc[rc, :] = np.interp(ds.freqs, sc.freqs, sc.Ites[rc])
        
    ds._Vth = 1./((1./Ites_ob - 1./Ites_sc) / ds._Rn[:, None])
    ds._Zeq = ds._Vth / Ites_sc

    ds._Ztes = ds._Vth / ds.Ites - ds._Zeq

    for f in fields:
        ds.move(f, f[1:])

    return ds


def Ztes_fit(f, R, beta, L, tau):
    """
    Ztes equation from Irwin/Shaw eq 42 
    """
    return R * (1 + beta) \
        + R * L / (1 - L) * (2 + beta) / (1 + 2j * np.pi * f * tau)

def guess_fit_params(ds, idx):
    """
    Gets initial params for fit at a particular freq idx
    """
    R = ds.Rtes[idx]

    min_idx = np.argmin(np.imag(ds.Ztes[idx]))
    tau_guess = -1./(2*np.pi*ds.freqs[min_idx])

    beta_guess = np.abs(ds.Ztes[idx, -1]) / R - 1

    L_guess = 1000

    return (R, beta_guess, L_guess, tau_guess)


def fit_single_det_params(ds, idx, x0=None, weights=None, fmax=None):
    """
    Fits detector parameters for a single channel
    """
    R = ds.Rtes[idx]
    if x0 is None:
        x0 = guess_fit_params(ds, idx)

    if weights is None:
        weights = np.ones_like(ds.freqs)

    if fmax is not None:
        weights[ds.freqs > fmax] = 0

    def chi2(x):
        zfit = Ztes_fit(ds.freqs, *x)
        c2 = np.nansum(weights * np.abs(ds.Ztes[idx] - zfit)**2)
        return c2

    res = minimize(chi2, x0)
    if list(res.x) == list(x0):
        res.success = False

    return res

def fit_det_params(ds, pb=False, fmax=None):
    """
    Fits detector params for a sweep.

    Args
    -----
    ds : AxisManager
        CIData
    pb : bool
        If True, wil display progressbar. 
    fmax : optional, float
        If set, will only fit using freq values less than fmax.
    """
    fields = ['_fit_x', '_fit_labels', '_tau_eff', '_Rfit', '_beta_I', '_L_I',
        '_tau_I']

    for f in fields:
        if f in ds._fields:
            ds.move(f, None)

    for f in ['_tau_eff', '_Rfit', '_beta_I', '_L_I', '_tau_I']:
        ds.wrap_new(f, ('dets', ), cls=np.full, fill_value=np.nan)
    ds.wrap('_fit_labels', np.array(['R', 'beta_I', 'L_I', 'tau_I']))
    ds.wrap_new('_fit_x', ('dets', 4), cls=np.full, fill_value=np.nan)

    for i in trange(len(ds.channels), disable=(not pb)):
        if ds.bgmap[i] == -1: continue

        res = fit_single_det_params(ds, i, fmax=fmax)
        ds._fit_x[i, :] = res.x
    
    # Compute tau_eff
    RL = ds.meta.R_sh
    ds._Rfit, ds._beta_I, ds._L_I, ds._tau_I = ds._fit_x.T
    ds._tau_eff = ds._tau_I * (1 - ds._L_I) * (1 + ds._beta_I + RL / ds._Rfit) \
        / (1 + ds._beta_I + RL / ds._Rfit + ds._L_I * (1 - RL / ds._Rfit))

    for f in fields:
        ds.move(f, f[1:])

    return True

def analyze_full(ds, bgs=None):
    """
    Performs the full CI analysis on a dataset.
    """
    analyze_tods(ds, bgs=bgs)
    if ds.run_kwargs.state == 'transition':
        get_ztes(ds)
        fit_det_params(ds)
    return ds


###########################################################################
# Plotting functions
###########################################################################
def plot_transfers(d, rc):
    """
    Plot the SC, OB, and in-transition transfer functions for a channel
    """
    bg = d.bgmap[rc]

    fig, axes = plt.subplots(1, 2, figsize=(16, 6))
    labels = ['Transition', 'superconducting', 'overbiased']

    for i, s in enumerate([d, d.sc, d.ob]):
        mag = np.abs(s.Ites[rc])/s.Ibias[bg]
        phase = np.unwrap(np.angle(s.Ites[rc]))
        axes[0].plot(s.freqs, mag, '.', label=labels[i])
        axes[1].plot(s.freqs, phase, '.', label=labels[i])
    axes[0].set(xscale='log', yscale='log')
    for ax in axes:
        ax.legend()
        ax.set_xlabel('Freq (Hz)', fontsize=16)
    axes[0].set_ylabel(r'$I_\mathrm{TES}$ / $I_\mathrm{bias}$', fontsize=16)
    axes[1].set_ylabel(r'$\phi_{I_\mathrm{TES}}$', fontsize=16)
    return fig, axes

def plot_ztes(ds, rc, x=None, write_text=True):
    """
    plots Ztes data

    Args
    -----
    ds : AxisManager
        analyzed CIData object
    rc : int
        Channel whose data to plot
    x : optional, list
        Fit params to plot instead of the ones stored in the CIData object.
    write_text : bool
        If textbox with param data should be written.
    """
    dims = np.array([2.5, 1])
    fig, axes = plt.subplots(1, 2, figsize=5 * dims)

    ztes = 1000 * ds.Ztes[rc]
    fs = np.linspace(0, np.max(ds.freqs), 1000)
    if x is None:
        x = ds.fit_x[rc]
    zfit = 1000 * Ztes_fit(fs, *x)

    # Circ plot
    ax = axes[0]
    ax.scatter(np.real(ztes), np.imag(ztes), c=np.log(ds.freqs), marker='.')
    ax.plot(np.real(zfit), np.imag(zfit), color='black', ls='--', alpha=0.6)
    ax.set_xlabel(r'Re[$Z_\mathrm{TES}$] (m$\Omega$)', fontsize=16)
    ax.set_ylabel(r'Im[$Z_\mathrm{TES}$] (m$\Omega$)', fontsize=16)

    ## Param summary
    txt = '\n'.join([
        r'$\tau_\mathrm{eff}$ = ' + f'{ds.tau_eff[rc]*1000:.2f} ms',
        r'$R_\mathrm{fit}$ = ' + f'{x[0]*1000:.2f} '+ r'm$\Omega$',
        r'$\beta_I$ = ' + f'{x[1]:.2f}',
        r'$\mathcal{L}_I$ = ' + f'{x[2]:.2f}',
        r'$\tau_I$ = ' + f'{x[3]*1000:.2f} ms',
        # r'$r^2$ = ' + f'{ds.ztes_rsquared[rc]:.2f}',
    ])
    if write_text:
        ax.text(0.35, 0.35, txt, transform=ax.transAxes, fontsize=12,
                bbox=dict(facecolor='white', alpha=0.8, edgecolor='black'))

    # Im / Re plot
    ax = axes[1]
    ax.plot(ds.freqs, np.real(ztes), '.', color='C0', label='Re')
    ax.plot(ds.freqs, np.imag(ztes), '.', color='C1', label='Im')
    ax.plot(fs, np.real(zfit), color='C0', alpha=0.8, ls='--')
    ax.plot(fs, np.imag(zfit), color='C1', alpha=0.8, ls='--')
    ax.set(xscale='log')
    ax.set_xlabel("Freq (Hz)", fontsize=16)
    ax.set_ylabel(r"$Z_\mathrm{TES}$ (m$\Omega$)", fontsize=16)
    ax.legend(fontsize=12)

    return fig, ax

###########################################################################
# Data Taking Functions
###########################################################################
@sdl.set_action()
def take_complex_impedance(
        S, cfg, bgs, freqs=None, state='transition', nperiods=500,
        max_meas_time=20., tickle_voltage=0.005, run_analysis=False):
    """
    Takes a complex impedance sweep. This will play sine waves on specified
    bias-groups over the current DC bias voltage. This returns a CISweep object.

    Args
    ----
    S : SmurfControl
        Pysmurf Instance
    cfg : DetConfig
        Det config instance
    bgs : array, int
        List of bias groups to run on
    freqs : array, optional
        List of frequencies to sweep over.
    state : str
        Current detector state. Must be 'ob', 'sc', or 'transition'
    nperiods : float
        Number of periods to measure for at each frequency. If the meas_time 
        ends up larger than ``max_meas_time``, ``max_meas_time`` will be used
        instead. This makes it so we don't spend unreasonably long amounts of
        time at higher freqs.
    max_meas_time : float
        Maximum amount of time to wait at any given frequency
    tickle_voltage : float
        Tickle amplitude in low-current-mode volts.
    run_analysis : bool
        Perform the full CI analysis and save the results.
    """
    if state not in ['ob', 'sc', 'transition']:
        raise ValueError("State must be 'ob', 'sc', or 'transition'")

    bgs = np.atleast_1d(bgs)

    if freqs is None:
        freqs = np.logspace(0, np.log10(2e3), 20)
    freqs = np.atleast_1d(freqs)

    run_kwargs = {k: v for k, v in locals().items() if k not in ['S', 'cfg']}

    # First, determine which bands and channels we'll try to run on.
    scale_array = np.array([S.get_amplitude_scale_array(b) for b in range(8)])
    bands, channels = np.where(scale_array > 0)

    # Main dataset
    ds = new_ci_dset(S, cfg, bands, channels, freqs, run_kwargs=run_kwargs)

    initial_ds_factor = S.get_downsample_factor()
    initial_filter_disable = S.get_filter_disable()

    pb = tqdm(total=len(freqs)*len(bgs), disable=False)
    try:
        sdl.set_current_mode(S, bgs, 1)
        tickle_voltage /= S.high_low_current_ratio

        init_biases = S.get_tes_bias_bipolar_array()
        for bg in bgs:
            m = ds.bgmap == bg
            channel_mask = ds.bands[m] * S.get_number_channels() + ds.channels[m]

            ds.sids[bg] = sdl.stream_g3_on(
                S, channel_mask=channel_mask, subtype='complex_impedance',
                downsample_factor=1, filter_disable=True
            )
            for j, freq in enumerate(freqs):
                meas_time = min(1./freq * nperiods, max_meas_time)
                S.log(f"Tickle with bg={bg}, freq={freq}")
                S.play_sine_tes(bg, tickle_voltage, freq)
                ds.start_times[bg, j] = time.time()
                time.sleep(meas_time)
                ds.stop_times[bg, j] = time.time()
                S.set_rtm_arb_waveform_enable(0)
                S.set_tes_bias_bipolar(bg, init_biases[bg])
                pb.update()
            sdl.stream_g3_off(S)
    finally:
        sdl.set_current_mode(S, bgs, 0)
        S.set_downsample_factor(initial_ds_factor)
        S.set_filter_disable(initial_filter_disable)
        sdl.stream_g3_off(S)

    fname = sdl.make_filename(S, f'ci_sweep_{state}.h5')
    ds.wrap('filepath', fname)
    ds.save(fname)
    S.pub.register_file(fname, 'ci', format='h5', plot=False)
    S.log(f"Saved unanalyzed datafile to {fname}")

    if run_analysis:
        analyze_full(ds, bgs=bgs)
        fname = sdl.make_filename(S, f'ci_sweep_{state}.h5')
        ds.filepath = fname
        ds.save(fname)
        S.pub.register_file(fname, 'ci', format='h5', plot=False)
        S.log(f"Saved analyzed datafile to {fname}")

    return ds

def take_complex_impedance_ob_sc(S, cfg, bgs, overbias_voltage=19.9,
                                 tes_bias=15.0, overbias_wait=5.0,
                                 cool_wait=30., run_analysis=True, **ci_kwargs):
    """
    Takes overbiased and superconducting complex impedance sweeps. These are
    required to analyze any in-transition sweeps.

    Args
    -----
    S : SmurfControl
        Pysmurf Instance
    cfg : DetConfig
        Det config instance
    bgs : array, int
        List of bias groups to run on
    overbias_voltage : float
        Voltage to use to overbias detectors
    tes_bias : float
        Voltage to set detectors to after overbiasing
    overbias_wait : float
        Time to wait at the overbias_voltage
    cool_wait : float
        Time to wait at the tes_bias after overbiasing
    run_analysis : bool
        Perform the full CI analysis and save the results.
    **ci_kwargs : 
        Any additional kwargs will be passed directly to the
        ``take_complex_impedance`` function.
    """
    bgs = np.atleast_1d(bgs)

    # Takes SC sweep
    for bg in bgs:
        S.set_tes_bias_bipolar(bg, 0)
    sc = take_complex_impedance(S, cfg, bgs, state='sc',
                                run_analysis=run_analysis, **ci_kwargs)

    S.overbias_tes_all(bias_groups=bgs, overbias_voltage=overbias_voltage,
                       tes_bias=tes_bias, overbias_wait=overbias_wait,
                       cool_wait=cool_wait)
    ob = take_complex_impedance(S, cfg, bgs, state='ob',
                                run_analysis=run_analysis, **ci_kwargs)

    cfg.dev.update_experiment({
        'complex_impedance_sc_path': sc.filepath,
        'complex_impedance_ob_path': ob.filepath
    }, update_file=True)

    return sc, ob



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

import os
from functools import wraps
try:
    import epics
    from pysmurf.client.util.pub import set_action
except Exception as e:
    # Just return base function regularly if can't import set_action
    def set_action(*args, **kwargs):
        def dec(func):
            return func
        return dec
    os.environ['NO_PYSMURF'] = 'true'


from sodetlib.util import *
from sodetlib.stream import *
from sodetlib import noise




import numpy as np
import os
import time
from scipy.signal import find_peaks
from scipy.optimize import curve_fit
import sodetlib as sdl
from sodetlib.legacy.analysis import squid_fit as sqf
import matplotlib.pyplot as plt
from tqdm.auto import tqdm, trange


def autocorr(wave):
    """
    Code to calculate the autocorrelation function of signal ``wave``.

    Args
    ----
    wave : float, ndarray
        signal to calculate autocorrelation on
    Returns
    -------
    lags : float, ndarray
        time/x-shift of signal relative to itself
    corrs : float, ndarray
        correlation coefficient at each lag
    """
    n = len(wave)
    lags = range(len(wave) // 2)
    corrs = np.zeros(len(lags))
    for ix, lag in enumerate(lags):
        y1 = wave[lag:]
        y2 = wave[: n - lag]
        corrs[ix] = np.corrcoef(y1, y2)[0, 1]
    return lags, corrs


def estimate_fit_parameters(phi, noisy_squid_curve, nharmonics_to_estimate=5,
                            min_acorr_dist_from_zero_frac=0.1):
    """
    Estimate rf-SQUID curve fit parameters to pass to ``fit_squid_curves`` as
    an initial guess for fitter.

    Args
    ----
    phi : ndarray
       Array of fixed flux ramp bias voltages.
    noisy_squid_curve : ndarray
       Array of rf-SQUID resonator frequencies, one for each flux ramp
       bias voltage.
    nharmonics_to_estimate : int, optional, default 5
       Number of harmonics of fundamental squid curve frequency to estimate.
    min_acorr_dist_from_zero_frac : float, optional, default 0.1
       Minimum distance to first peak in autocorrelation (used to estimate what
       how much fr bias current corresponds to 1 phi0) in units fraction of
       the input ``phi`` array (between 0 to 1)

    Returns
    -------
    est : ndarray or None
       Estimated fit parameters. Returns None if unable to estimate
       the phi0 using lag and autocorrelation.
    """
    min_acorr_dist_from_zero = len(phi) * min_acorr_dist_from_zero_frac

    # find period from autocorrelation
    lags, corrs = autocorr(noisy_squid_curve)

    # find peaks in autocorrelation vs lag
    peaks, _ = find_peaks(corrs, height=0)
    sorted_peaks = sorted(peaks)

    try:
        phi0_idx = next(pk for pk in sorted_peaks if pk >
                        min_acorr_dist_from_zero)
    except:
        return None

    phi0 = np.abs(phi[phi0_idx] - phi[0])

    # plot cosine with same amplitude and period
    yspan = np.ptp(noisy_squid_curve)
    yoffset = yspan / 2.0 + np.min(noisy_squid_curve)

    def harmonic(n, ph, phoff, amp): return (amp) *\
        np.cos(n*(ph-phoff)*(2*np.pi/phi0))

    def first_harmonic_guess(ph, phoff): return yoffset + \
        harmonic(1, ph, phoff, (yspan / 2))

    # now correlate the first harmonic guess against the SQUID curve
    dphi = np.abs(phi[1] - phi[0])
    testphoffs = np.linspace(0, phi0, int(np.floor(phi0 / dphi) + 1))
    corrs = []
    for testphoff in testphoffs:
        y1 = first_harmonic_guess(phi, testphoff)
        y2 = noisy_squid_curve
        y1 = y1 - np.nanmean(y1)
        y2 = y2 - np.nanmean(y2)
        corr = np.corrcoef(y1, y2)[0, 1]
        corrs.append(corr)

    # should just be able to find the maximum of this correlation
    if np.isnan(corrs).all():
        phioffset = 0
    else:
        phioffset = testphoffs[np.nanargmax(corrs)]

    # plot harmonics only over the largest possible number of SQUID periods.  May only be 1.
    lower_phi_full_cycles = (np.min(phi) + phioffset) % (phi0) + np.min(phi)
    upper_phi_full_cycles = np.max(phi) - (np.max(phi) - phioffset) % phi0
    phi_full_cycle_idxs = np.where(
        (phi > lower_phi_full_cycles) & (phi < upper_phi_full_cycles)
    )
    phi_full_cycles = phi[phi_full_cycle_idxs]
    # correlate some harmonics and overplot!
    fit_guess = np.zeros_like(noisy_squid_curve) + np.nanmean(noisy_squid_curve)
    # mean subtract the data and this harmonic
    d = noisy_squid_curve[phi_full_cycle_idxs]
    dm = np.nanmean(d)
    d_ms = d - dm

    est = [phi0, phioffset, dm]

    for n in range(1, nharmonics_to_estimate + 1):
        def this_harmonic(ph): return harmonic(n, ph, phioffset, 1/2)
        h = this_harmonic(phi_full_cycles)
        hm = np.nanmean(h)
        h_ms = h - hm
        # sort of inverse dft them
        Xh = np.sum(d_ms * h_ms)
        # add this harmonic
        fit_guess += Xh * this_harmonic(phi)
        est.append(Xh)

    # match span of harmonic guess sum and add offset from data
    normalization_factor = np.ptp(d_ms) / np.ptp(fit_guess)
    fit_guess *= normalization_factor
    # also scale parameter guesses we pass back
    est = np.array(est)
    est[3:] *= normalization_factor
    return est


def squid_curve_model(phi, *p):
    """
    Functional form of squid curve, basically a fourier expansion (with
    n-harmonics).

    Args
    ----
    phi : float, ndarray
        depedent variable in function
    p : ndarray
        indices::
            0: period in fraction_full_scale
            1: offset phase in fraction_full_scale
            2: central value (mean) of squid curve
            3 to N: amplitude of the (N-3)th harmonic of the squid curve.
    """
    phi0 = p[0]
    phioffset = p[1]
    dm = p[2]
    ret = dm

    def harmonic(n, ph, phoff, amp): return (amp) *\
        np.cos(n*(ph-phoff)*(2*np.pi/phi0))
    for n in range(0, len(p[3:])):
        def this_harmonic(ph): return harmonic(n+1, ph, phioffset, 0.5*p[3+n])
        ret += this_harmonic(phi)
    return ret


def plot_squid_fit_summary(data, nbins=35, quantile=.95):
    '''
    Makes summary plots of squid curve fit parameters. 
    Specifically, Phi_offset, df, dfdI, hhpwr, phi0.

    Args
    ----
    data : dict
        Dictionary output created from take_squid_curves.
        Used to find data['bands'].size and data['popts'] values.
    nbins : int
        Number of bins to use for histograms.
    quantile : float
        Plot data that falls within the quantile given.
    '''
#     Plot offsets
    popts = data['popts']
    fig = plt.figure(figsize=(20, 15))
    fig.set_facecolor('white')
    offsets = popts[:, 1] / popts[:, 0]
    offsets = offsets[(offsets >= np.nanquantile(offsets, 1-quantile))
                      & (offsets <= np.nanquantile(offsets, quantile))]
    plt.subplot(3, 2, 1)
    plt.hist(offsets, bins=nbins, alpha=0.7)
    plt.axvline(np.median(offsets), color='C1', ls=':', lw=3)
    plt.legend([f'Median : {np.round(np.median(offsets),3)}', f'Data (Inner {quantile*100}%)\n {offsets.size} / {data["bands"].size} channels used'],
               fontsize=14)
    plt.xlabel('$\\Phi_0$ Offset', fontsize=16)
    plt.ylabel('Counts', fontsize=16)
    plt.tick_params(axis='both', which='major', labelsize=12)

#     Plot dfs
    dfs = data['df']
    dfs = dfs[(dfs >= np.nanquantile(dfs, 1-quantile))
              & (dfs <= np.nanquantile(dfs, quantile))]
    plt.subplot(3, 2, 2)
    plt.hist(dfs, bins=nbins, alpha=0.7)
    plt.axvline(np.median(dfs), color='C1', ls=':', lw=3)
    plt.legend([f'Median : {np.round(np.median(dfs),1)}', f'Data (Inner {quantile*100}%)\n {dfs.size} / {data["bands"].size} channels used'],
               fontsize=14)
    plt.xlabel('$df_{pp}$ [kHz]', fontsize=16)
    plt.ylabel('Counts', fontsize=16)
    plt.tick_params(axis='both', which='major', labelsize=12)

#     Plot dfdIs
    dfdIs = data['dfdI']
    dfdIs = dfdIs[(dfdIs >= np.nanquantile(dfdIs, 1-quantile))
                  & (dfdIs <= np.nanquantile(dfdIs, quantile))]
    plt.subplot(3, 2, 3)
    plt.hist(dfdIs/1e-3, bins=nbins, alpha=0.7)
    plt.axvline(np.median(dfdIs)/1e-3, color='C1', ls=':', lw=3)
    plt.legend([f'Median : {np.round(np.median(dfdIs)/1e-3,2)}', f'Data (Inner {quantile*100}%)\n {dfdIs.size} / {data["bands"].size} channels used'],
               fontsize=14)
    plt.xlabel('<df/dI> [mHz/pA]', fontsize=16)
    plt.ylabel('Counts', fontsize=16)
    plt.tick_params(axis='both', which='major', labelsize=12)

#     Plot hhpwrs
    hhpwrs = data['higher_harmonic_power']
    hhpwrs = hhpwrs[(hhpwrs >= np.nanquantile(hhpwrs, 1-quantile))
                    & (hhpwrs <= np.nanquantile(hhpwrs, quantile))]
    plt.subplot(3, 2, 4)
    plt.hist(hhpwrs*100, bins=nbins, alpha=0.7)
    plt.axvline(np.median(hhpwrs)*100, color='C1', ls=':', lw=3)
    plt.legend([f'Median : {np.round(np.median(hhpwrs)*100,2)}', f'Data (Inner {quantile*100}%)\n {hhpwrs.size} / {data["bands"].size} channels used'],
               fontsize=14)
    plt.xlabel('Higher Harmonic Power [%]', fontsize=16)
    plt.ylabel('Counts', fontsize=16)
    plt.tick_params(axis='both', which='major', labelsize=12)

#     Plot phio0s
    phi0s = data['ffs_per_phi0']
    phi0s = phi0s[(phi0s >= np.nanquantile(phi0s, 1-quantile))
                  & (phi0s <= np.nanquantile(phi0s, quantile))]
    plt.subplot(3, 2, 5)
    plt.hist(phi0s, bins=nbins, alpha=0.7)
    plt.axvline(np.median(phi0s), color='C1', ls=':', lw=3)
    plt.legend([f'Median : {np.round(np.median(phi0s),3)}', f'Data (Inner {quantile*100}%)\n {phi0s.size} / {data["bands"].size} channels used'],
               fontsize=14)
    plt.xlabel('$\\Phi_0$ [ff]', fontsize=16)
    plt.ylabel('Counts', fontsize=16)
    plt.tight_layout()
    plt.tick_params(axis='both', which='major', labelsize=12)
    return


def plot_squid_fit(data, band, channel, save_plot=False, S=None,
                   plot_dir=None):
    """
    Plots data taken with ``take_squid_curve`` against fits from
    ``fit_squid_curve``.

    Args
    ----
    data: dict
        Output data product from take_squid_curves
    band: int
        Band number
    channel: int
        Channel number
    save_plot: bool
        Boolean to toggle saving of plot to disk
    S: PysmurfControl object, optional
        Used to grab plot_dir and publish plots
    plot_dir: str
        Overrides plot_dir from S, if specified
    """
    if save_plot:
        if plot_dir is not None:
            pass
        elif S is not None:
            plot_dir = S.plot_dir
        else:
            raise ValueError("Either S or ``plot_dir`` must be specified.")
    idx = np.where((data['bands'] == band) & (
        data['channels'] == channel))[0][0]
    biases = data['fluxramp_ffs']

    fit_biases = np.linspace(np.min(biases), np.max(biases), 1000)

    fres = data['res_freq'][idx]
    fig = plt.figure()
    fig.set_facecolor('white')
    plt.plot(biases, data["res_freq_vs_fr"][idx, :], 'co')
    plt.plot(fit_biases, squid_curve_model(fit_biases,
                                       *data['popts'][idx, :]),
             'C1--')
    ax = plt.gca()
    plt.text(0.0175, 0.975, data['plt_txt'][idx], horizontalalignment='left',
             verticalalignment='top', transform=ax.transAxes, fontsize=10,
             bbox=dict(facecolor='wheat', alpha=0.5, boxstyle='round'))
    plt.xlabel("Flux Bias [Fraction Full Scale FR DAC]", fontsize=14)
    plt.ylabel("Frequency Swing [MHz]", fontsize=14)
    plt.title(
        f"Band {band} Channel {channel} $f_r$ = {np.round(fres,2)} [MHz]")
    if save_plot:
        ctime = int(time.time())
        fig_name = f"{plot_dir}/{ctime}_b{band}c{channel}_dc_squid_curve.png"
        plt.savefig(fig_name)
        if S is not None:
            S.pub.register_file(fig_name, 'dc_squid_curve', plot=True)
    return


def fit_squid_curves(squid_data, fit_args=None, modify_data=True, show_pb=False):
    '''
    Function for fitting squid curves taken with ``take_squid_curve``.

    Args
    ----
    squid_data: dictionary
        Output data product from take_squid_curves
    fit_args: dictionary
        Dictionary of keyword arguments to be passed to 
        estimate_fit_parameters

    Returns
    -------
    fit_out: dictionary
        Dictionary of fit parameters.
    '''
    fit_out = {}

    if fit_args is None:
        fit_args = {}
    nharm = fit_args.setdefault('nharmonics_to_estimate', 5)

    nchans = len(squid_data['channels'])
    fit_guess = np.zeros((nchans, nharm+3))
    fit_result = np.zeros((nchans, nharm+3))
    plt_txt = []
    df = np.zeros(nchans)
    dfdI = np.zeros(nchans)
    hhpwr = np.zeros(nchans)
    for nc in trange(nchans, disable=not show_pb):
        fit_guess[nc, :] = estimate_fit_parameters(squid_data['fluxramp_ffs'],
                                                   squid_data['res_freq_vs_fr'][nc],
                                                   **fit_args)
        fit_result[nc, :], _ = curve_fit(squid_curve_model, squid_data['fluxramp_ffs'],
                                         squid_data['res_freq_vs_fr'][nc, :], p0=fit_guess[nc, :])

        txt, params = get_derived_params_and_text(
            squid_data, fit_result[nc, :], nc)
        df[nc] = params['df']
        dfdI[nc] = params['dfdI']
        hhpwr[nc] = params['higher_harmonic_power']
        plt_txt.append(txt)

    fit_out['initial_guess'] = fit_guess
    fit_out['model_params'] = fit_result
    fit_out['df'] = df
    fit_out['dfdI'] = dfdI
    fit_out['higher_harmonic_power'] = hhpwr
    fit_out['plt_txt'] = np.array(plt_txt)

    if modify_data:
        squid_data.update(fit_out)
        squid_data['popts'] = fit_result
        squid_data['ffs_per_phi0'] = fit_result[:, 0]

    return fit_out


def dfduPhi0_to_dfdI(dfduphi0, M_in=227e-12):
    '''
    Function to convert averaged squid slope (average squid gain) from units of
    Hz/micro-Phi0 to H per pA given the mutual inductance between the squid and
    TES circuit.

    Args
    ----
    dfduphi0 : float
        SQUID gain in Hz per micro-Phi0
    M_in : float
        Mutual inductance between SQUID and TES in units of Henries.
    Returns
    -------
    dfdI_Hz_per_pA : float
        SQUID gain in Hz per pA
    '''
    M_in_phi0_per_A = M_in/2.067833848e-15  # Phi0/A = [Wb/A]/[Wb/Phi0]
    dfdphi0 = dfduphi0*1e6  # Hz/phi0 = [Hz/uphi0]*[uphi0/phi0]
    dfdI_Hz_per_A = dfdphi0*M_in_phi0_per_A  # Hz/A = [HZ/Phi0]*[Phi0/A]
    dfdI_Hz_per_pA = dfdI_Hz_per_A*1e-12  # Hz/pA = [Hz/A]*[A/pA]
    return dfdI_Hz_per_pA


def get_derived_params_and_text(data, model_params, idx):
    '''
    Function to calculate some useful parameters derived from the primary squid
    curve fit parameters and return a text block to add to channel plots.

    Args
    ----
    data : ndarray
        Output data dictionary from ``take_squid_curve``.
    model_params : dict
        Model parameters fitted to ``squid_curve_model``.
    idx : int
        Index of output data array for channel to get text for.
    Returns
    -------
    plot_txt : str
        String to add to squid fit channel plot.
    derived_params : dict
        Dictionary of useful parameters derived from the model fit.
    '''
    fit_curve = sqf.model(data['fluxramp_ffs'], *model_params)
    df_khz = np.ptp(fit_curve)*1000
    phi_over_one_cycle = np.linspace(0, model_params[0], 10000)+model_params[1]
    fit_curve_over_one_cycle = squid_curve_model(
        phi_over_one_cycle, *model_params)
    phi_over_one_cycle /= model_params[0]
    avg_dfdphi_Hzperuphi0 = np.nanmean(np.abs(np.gradient(fit_curve_over_one_cycle)) /
                                    (np.gradient(phi_over_one_cycle)))
    hhpwr = (np.square(model_params[3]) /
             (np.sum(np.square(model_params[4:]))))**-1
    dfdI = dfduPhi0_to_dfdI(avg_dfdphi_Hzperuphi0, 227e-12)
    plot_txt = ('$f_{res}$' + f' = {data["res_freq"][idx]:.1f} MHz\n' +
                '$\\Phi_0$' + f' = {model_params[0]:.3f} ff\n' +
                '$\\Phi_{offset}$' +
                f' = {model_params[1]/model_params[0]:.3f} Phi0\n' +
                f'df = {df_khz:.1f} kHz\n' +
                '$<df/dI>$' + f' = {dfdI/1e-3:.1f} mHz/pA\n' +
                f'hhpwr = {hhpwr:.3f}')
    derived_params = {'df': df_khz,
                      'dfdI': dfdI,
                      'higher_harmonic_power': hhpwr}
    return plot_txt, derived_params


@sdl.set_action()
def take_squid_curve(S, cfg, wait_time=0.1, Npts=4, Nsteps=500,
                     bands=None, channels=None, lms_gain=None, out_path=None,
                     run_analysis=True, analysis_kwargs=None, show_pb=False,
                     run_serial_ops=True, frac_full_scale_max=0.3):
    """
    Takes data in open loop (only slow integral tracking) and steps through flux
    values to trace out a SQUID curve. This can be compared against the tracked
    SQUID curve which might not perfectly replicate this if these curves are
    poorly approximated by a sine wave (or ~3 harmonics of a fourier expansion).

    Args
    ----
    S : ``pysmurf.client.base.smurf_control.SmurfControl``
        ``pysmurf`` control object
    cfg : ``sodetlib.det_config.DeviceConfig``
        device config object.
    wait_time: float
        how long you wait between flux step point in seconds
    Npts : int
        number of points you take at each flux bias step to average
    Nsteps : int
        Number of flux points you will take total.
    bands : int, list
        list of bands to take dc SQUID curves on
    channels : dict
        default is None and will run on all channels that are on
        otherwise pass a dictionary with a key for each band
        with values equal to the list of channels to run in each band.
    lms_gain : int
        gain used in tracking loop filter and set in ``tracking_setup``
        defaults to ``None`` and pulls from ``det_config``
    out_path : str, filepath
        directory to output npy file to. defaults to ``None`` and uses pysmurf
        plot directory (``S.plot_dir``)
    run_serial_ops : bool
        If true, will run serial grad descent and eta scan
    frac_full_scale_max : float
        Max value of fraction full scale to use.
    Returns
    -------
    data : dict
        This contains the flux bias array, channel array, and frequency
        shift at each bias value for each channel in each band. As well as
        the dictionary of fitted values returned by fit_squid_curves if 
        run_analysis argument is set to True.
    """
    cur_mode = S.get_cryo_card_ac_dc_mode()
    if cur_mode == 'AC':
        S.set_mode_dc()
    ctime = S.get_timestamp()
    if out_path is None:
        out_path = os.path.join(S.output_dir, f'{ctime}_fr_sweep_data.npy')

    # This calculates the amount of flux ramp amplitude you need for 1 phi0
    # and then sets the range of flux bias to be enough to achieve the Nphi0s
    # specified in the fucnction call.
    if bands is None:
        bands = cfg.dev.exp['active_bands']
    bands = np.atleast_1d(bands)

    if channels is None:
        channels = {}
        for band in bands:
            channels[band] = S.which_on(band)

    bias_peak = frac_full_scale_max

    # This is the step size calculated from range and number of steps
    bias_step = np.abs(2*bias_peak)/float(Nsteps)

    channels_out = []
    bands_out = []
    for band in bands:
        channels_out.extend(channels[band])
        bands_out.extend(list(np.ones(len(channels[band]))*band))
    biases = np.arange(-bias_peak, bias_peak, bias_step)

    # final output data dictionary
    data = {}
    data['meta'] = sdl.get_metadata(S, cfg)
    data['bands'] = np.asarray(bands_out)
    data['channels'] = np.asarray(channels_out)
    data['fluxramp_ffs'] = biases
    data['res_freq_vs_fr'] = []

    unique_bands = np.unique(np.asarray(bands_out, dtype=int))
    prev_lms_enable1 = {}
    prev_lms_enable2 = {}
    prev_lms_enable3 = {}
    prev_lms_gain = {}

    # take SQUID data
    try:
        for band in unique_bands:
            band_cfg = cfg.dev.bands[band]
            if lms_gain is None:
                lms_gain = band_cfg['lms_gain']
            S.log(f'{len(channels[band])} channels on in band {band},'
                  ' configuring band for simple, integral tracking')
            S.log(
                f'-> Setting lmsEnable[1-3] and lmsGain to 0 for band {band}.')
            prev_lms_enable1[band] = S.get_lms_enable1(band)
            prev_lms_enable2[band] = S.get_lms_enable2(band)
            prev_lms_enable3[band] = S.get_lms_enable3(band)
            prev_lms_gain[band] = S.get_lms_gain(band)
            S.set_lms_enable1(band, 0)
            S.set_lms_enable2(band, 0)
            S.set_lms_enable3(band, 0)
            S.set_lms_gain(band, lms_gain)

        fs = {}
        S.log(
            '\rSetting flux ramp bias to 0 V\033[K before tune'.format(-bias_peak))
        S.set_fixed_flux_ramp_bias(0.)

        for band in unique_bands:
            fs[band] = []
            if run_serial_ops:
                S.run_serial_gradient_descent(band)
                S.run_serial_eta_scan(band)
            S.toggle_feedback(band)

        small_steps_to_starting_bias = np.arange(
            -bias_peak, 0, bias_step)[::-1]

        # step from zero (where we tuned) down to starting bias
        S.log('Slowly shift flux ramp voltage to place where we begin.')
        for b in small_steps_to_starting_bias:
            S.set_fixed_flux_ramp_bias(b, do_config=False)
            time.sleep(wait_time)

        # make sure we start at bias_low
        S.log(f'\rSetting flux ramp bias low at {-bias_peak} V')
        S.set_fixed_flux_ramp_bias(-bias_peak, do_config=False)
        time.sleep(wait_time)

        S.log('Starting to take flux ramp.')

        for b in tqdm(biases, disable=(not show_pb)):
            S.set_fixed_flux_ramp_bias(b, do_config=False)
            time.sleep(wait_time)
            for band in unique_bands:
                fsamp = np.zeros(shape=(Npts, len(channels[band])))
                for i in range(Npts):
                    fsamp[i, :] = S.get_loop_filter_output_array(band)[
                        channels[band]]
                fsampmean = np.nanmean(fsamp, axis=0)
                fs[band].append(fsampmean)

        S.log('Done taking flux ramp data.')
        fres = []
        for i, band in enumerate(unique_bands):
            fres_loop = S.channel_to_freq(band).tolist()
            fres.extend(fres_loop)
            # stack
            lfovsfr = np.dstack(fs[band])[0]
            fvsfr = np.array(
                [arr+fres for (arr, fres) in zip(lfovsfr, fres_loop)])
            if i == 0:
                data['res_freq_vs_fr'] = fvsfr
            else:
                data['res_freq_vs_fr'] = np.concatenate(
                    (data['res_freq_vs_fr'], fvsfr), axis=0)

        data['res_freq'] = np.asarray(fres)
        data['filepath'] = out_path

    # always zero and restore state of system
    finally:
        # done - zero and unset
        S.set_fixed_flux_ramp_bias(0, do_config=False)
        S.unset_fixed_flux_ramp_bias()
        for band in unique_bands:
            S.set_lms_enable1(band, prev_lms_enable1[band])
            S.set_lms_enable2(band, prev_lms_enable2[band])
            S.set_lms_enable3(band, prev_lms_enable3[band])
            S.set_lms_gain(band, lms_gain)

        if cur_mode == 'AC':
            S.set_mode_ac()

    if run_analysis:
        kw = {'modify_data': True}
        if analysis_kwargs is not None:
            kw.update(analysis_kwargs)
        try:
            fit_dict = fit_squid_curves(data, **kw)
        except Exception as e:
            S.log("Analysis of SQUID Curve failed!")


    # save dataset for each iteration, just to make sure it gets
    # written to disk
    np.save(out_path, data)
    S.pub.register_file(out_path, 'dc_squid_curve', format='npy')

    return data



import pytest
from sodetlib.det_config import DetConfig
import os


@pytest.fixture
def cfg():
    (sys, dev, pysmurf) = [
        os.path.join(os.path.dirname(__file__), f'input/{fname}')
        for fname in ['sys_config.yml', 'dev_cfg.yml', 'pysmurf.cfg']
    ]
    return DetConfig(sys_file=sys, dev_file=dev, pysmurf_file=pysmurf)


def test_dev_cfg(cfg):
    cfg.parse_args(args=['-N', '2'])
    cfg.dev.update_band(1, {'dc_att': 3, 'drive': 11})
    cfg.dev.update_bias_group(0, {'bias_high': 11, 'enabled': 1})
    cfg.dev.update_experiment({"amp_50k_Id": 12})
    out = cfg.dump_configs('config', clobber=True)
    sys, dev, pysmurf = out['sys'], out['dev'], out['pysmurf']
    # Testing reloading written config file
    cfg = DetConfig(sys_file=sys, dev_file=dev, pysmurf_file=pysmurf)
    cfg.parse_args(args=['-N', '2'])
    assert (cfg.dev.bands[1]['dc_att'] == 3)


def test_failed_update(cfg):
    cfg.parse_args(args=[])
    with pytest.raises(ValueError):
        cfg.dev.update_experiment({'abcd': 12})


def test_uxm_file(cfg):
    cfg.uxm_file= 'input/uxm.yaml'
    cfg.parse_args(args=['-N', '2'])
    assert (cfg.uxm is not None)


def test_offline_pysmurf_instance(cfg):
    cfg.parse_args(args=[])
    cfg.get_smurf_control(offline=True, dump_configs=True)


def test_online_dump_confg():
    cfg = DetConfig()
    cfg.load_config_files(slot=2)
    cfg.get_smurf_control(dump_configs=False)
    config_dir = 'config_dump'
    cfg.dump_configs(config_dir, clobber=True, dump_rogue_tree=True)
